"""Tool and workflow learning primitives for adaptive runtime behavior."""
from __future__ import annotations

from dataclasses import dataclass, field
import math
import os
from pathlib import Path
import sqlite3
from typing import Any

from logos_adk.persistence import (
    get_postgresql_schema_version,
    get_sqlite_schema_version,
    set_postgresql_schema_version,
    set_sqlite_schema_version,
)
from logos_adk.postgres_utils import import_psycopg, resolve_database_url

from .models import utcnow


TOOL_WORKFLOW_LEARNING_SCHEMA_KIND = "tool_workflow_learning"
TOOL_WORKFLOW_LEARNING_SCHEMA_VERSION = 1


def _segment(value: str | None) -> str:
    return value or ""


def _score(successes: int, failures: int) -> float:
    total = successes + failures
    if total <= 0:
        return 0.5
    return (successes + 1.0) / (total + 2.0)


@dataclass
class ToolScore:
    """Aggregated tool performance for a task segment."""

    tool_name: str
    task_type: str | None = None
    workspace_id: str | None = None
    success_count: int = 0
    failure_count: int = 0
    average_duration_ms: float = 0.0
    win_rate: float = 0.0
    score: float = 0.0
    last_updated: str = field(default_factory=utcnow)


@dataclass
class FallbackPolicyScore:
    """Aggregated fallback provider effectiveness for a task segment."""

    provider_key: str
    task_type: str | None = None
    workspace_id: str | None = None
    success_count: int = 0
    failure_count: int = 0
    average_duration_ms: float = 0.0
    win_rate: float = 0.0
    score: float = 0.0
    last_updated: str = field(default_factory=utcnow)


@dataclass
class WorkflowBranchScore:
    """Aggregated workflow branch effectiveness."""

    pipeline_name: str
    step_name: str
    branch_name: str
    workspace_id: str | None = None
    success_count: int = 0
    failure_count: int = 0
    average_duration_ms: float = 0.0
    score: float = 0.0
    last_updated: str = field(default_factory=utcnow)


@dataclass
class ApprovalScore:
    """Aggregated approval friction for a step."""

    pipeline_name: str
    step_name: str
    workspace_id: str | None = None
    request_count: int = 0
    approval_count: int = 0
    rejection_count: int = 0
    correction_rate: float = 0.0
    approval_rate: float = 0.0
    last_updated: str = field(default_factory=utcnow)


@dataclass
class LoopRecommendation:
    """Recommended loop limit based on historical executions."""

    pipeline_name: str
    step_name: str
    workspace_id: str | None = None
    recommended_limit: int = 1
    average_iterations: float = 0.0
    max_observed_iterations: int = 0
    max_hits: int = 0
    sample_size: int = 0
    last_updated: str = field(default_factory=utcnow)


class ToolWorkflowLearningEngine:
    """Persistent learning engine for tool and workflow recommendations."""

    def __init__(
        self,
        path: str | None = None,
        table_prefix: str | None = None,
        dsn: str | None = None,
    ) -> None:
        self.path = path or os.getenv(
            "LOGOS_ADK_TOOL_WORKFLOW_DB_PATH",
            ".logos_adk/tool_workflow_learning.db",
        )
        self.dsn = resolve_database_url(
            dsn or os.getenv("LOGOS_ADK_TOOL_WORKFLOW_DSN"),
            env_vars=("LOGOS_ADK_TOOL_WORKFLOW_DSN", "LOGOS_ADK_DATABASE_URL", "DATABASE_URL"),
        )
        self.table_prefix = table_prefix or os.getenv(
            "LOGOS_ADK_TOOL_WORKFLOW_TABLE_PREFIX",
            "tool_workflow_learning",
        )

    def _table(self, suffix: str) -> str:
        return f"{self.table_prefix}_{suffix}"

    def _connect(self) -> sqlite3.Connection | Any:
        if self.dsn:
            connection = import_psycopg().connect(self.dsn)
            self._ensure_schema(connection)
            return connection
        db_path = Path(self.path)
        if db_path.parent and not db_path.parent.exists():
            db_path.parent.mkdir(parents=True, exist_ok=True)
        connection = sqlite3.connect(self.path)
        connection.row_factory = sqlite3.Row
        self._ensure_schema(connection)
        return connection

    def _ensure_schema(self, connection: sqlite3.Connection | Any) -> None:
        if self.dsn:
            current = get_postgresql_schema_version(connection, TOOL_WORKFLOW_LEARNING_SCHEMA_KIND) or 0
            with connection.cursor() as cursor:
                cursor.execute(
                    f"""
                    CREATE TABLE IF NOT EXISTS {self._table('tool_events')} (
                        id TEXT PRIMARY KEY,
                        workspace_id TEXT NOT NULL,
                        task_type TEXT NOT NULL,
                        tool_name TEXT NOT NULL,
                        success INTEGER NOT NULL,
                        duration_ms DOUBLE PRECISION NOT NULL,
                        error TEXT,
                        trace_id TEXT,
                        created_at TEXT NOT NULL
                    )
                    """
                )
                cursor.execute(
                    f"""
                    CREATE TABLE IF NOT EXISTS {self._table('fallback_events')} (
                        id TEXT PRIMARY KEY,
                        workspace_id TEXT NOT NULL,
                        task_type TEXT NOT NULL,
                        provider_key TEXT NOT NULL,
                        success INTEGER NOT NULL,
                        duration_ms DOUBLE PRECISION NOT NULL,
                        was_primary INTEGER NOT NULL DEFAULT 0,
                        error TEXT,
                        trace_id TEXT,
                        created_at TEXT NOT NULL
                    )
                    """
                )
                cursor.execute(
                    f"""
                    CREATE TABLE IF NOT EXISTS {self._table('workflow_branch_events')} (
                        id TEXT PRIMARY KEY,
                        workspace_id TEXT NOT NULL,
                        pipeline_name TEXT NOT NULL,
                        step_name TEXT NOT NULL,
                        branch_name TEXT NOT NULL,
                        success INTEGER NOT NULL,
                        duration_ms DOUBLE PRECISION NOT NULL,
                        trace_id TEXT,
                        created_at TEXT NOT NULL
                    )
                    """
                )
                cursor.execute(
                    f"""
                    CREATE TABLE IF NOT EXISTS {self._table('loop_events')} (
                        id TEXT PRIMARY KEY,
                        workspace_id TEXT NOT NULL,
                        pipeline_name TEXT NOT NULL,
                        step_name TEXT NOT NULL,
                        iterations INTEGER NOT NULL,
                        max_iterations INTEGER NOT NULL,
                        termination_reason TEXT NOT NULL,
                        success INTEGER NOT NULL,
                        trace_id TEXT,
                        created_at TEXT NOT NULL
                    )
                    """
                )
                cursor.execute(
                    f"""
                    CREATE TABLE IF NOT EXISTS {self._table('approval_events')} (
                        id TEXT PRIMARY KEY,
                        workspace_id TEXT NOT NULL,
                        pipeline_name TEXT NOT NULL,
                        step_name TEXT NOT NULL,
                        event_type TEXT NOT NULL,
                        approved INTEGER,
                        corrected INTEGER,
                        trace_id TEXT,
                        created_at TEXT NOT NULL
                    )
                    """
                )
                cursor.execute(
                    f"CREATE INDEX IF NOT EXISTS idx_{self._table('tool_events')}_segment ON {self._table('tool_events')} (workspace_id, task_type, tool_name, created_at DESC)"
                )
                cursor.execute(
                    f"CREATE INDEX IF NOT EXISTS idx_{self._table('fallback_events')}_segment ON {self._table('fallback_events')} (workspace_id, task_type, provider_key, created_at DESC)"
                )
                cursor.execute(
                    f"CREATE INDEX IF NOT EXISTS idx_{self._table('workflow_branch_events')}_segment ON {self._table('workflow_branch_events')} (workspace_id, pipeline_name, step_name, branch_name, created_at DESC)"
                )
                cursor.execute(
                    f"CREATE INDEX IF NOT EXISTS idx_{self._table('loop_events')}_segment ON {self._table('loop_events')} (workspace_id, pipeline_name, step_name, created_at DESC)"
                )
                cursor.execute(
                    f"CREATE INDEX IF NOT EXISTS idx_{self._table('approval_events')}_segment ON {self._table('approval_events')} (workspace_id, pipeline_name, step_name, created_at DESC)"
                )
            if current < TOOL_WORKFLOW_LEARNING_SCHEMA_VERSION:
                set_postgresql_schema_version(
                    connection,
                    TOOL_WORKFLOW_LEARNING_SCHEMA_KIND,
                    TOOL_WORKFLOW_LEARNING_SCHEMA_VERSION,
                )
            connection.commit()
            return

        current = get_sqlite_schema_version(connection, TOOL_WORKFLOW_LEARNING_SCHEMA_KIND) or 0
        connection.execute(
            f"""
            CREATE TABLE IF NOT EXISTS {self._table('tool_events')} (
                id TEXT PRIMARY KEY,
                workspace_id TEXT NOT NULL,
                task_type TEXT NOT NULL,
                tool_name TEXT NOT NULL,
                success INTEGER NOT NULL,
                duration_ms REAL NOT NULL,
                error TEXT,
                trace_id TEXT,
                created_at TEXT NOT NULL
            )
            """
        )
        connection.execute(
            f"""
            CREATE TABLE IF NOT EXISTS {self._table('fallback_events')} (
                id TEXT PRIMARY KEY,
                workspace_id TEXT NOT NULL,
                task_type TEXT NOT NULL,
                provider_key TEXT NOT NULL,
                success INTEGER NOT NULL,
                duration_ms REAL NOT NULL,
                was_primary INTEGER NOT NULL DEFAULT 0,
                error TEXT,
                trace_id TEXT,
                created_at TEXT NOT NULL
            )
            """
        )
        connection.execute(
            f"""
            CREATE TABLE IF NOT EXISTS {self._table('workflow_branch_events')} (
                id TEXT PRIMARY KEY,
                workspace_id TEXT NOT NULL,
                pipeline_name TEXT NOT NULL,
                step_name TEXT NOT NULL,
                branch_name TEXT NOT NULL,
                success INTEGER NOT NULL,
                duration_ms REAL NOT NULL,
                trace_id TEXT,
                created_at TEXT NOT NULL
            )
            """
        )
        connection.execute(
            f"""
            CREATE TABLE IF NOT EXISTS {self._table('loop_events')} (
                id TEXT PRIMARY KEY,
                workspace_id TEXT NOT NULL,
                pipeline_name TEXT NOT NULL,
                step_name TEXT NOT NULL,
                iterations INTEGER NOT NULL,
                max_iterations INTEGER NOT NULL,
                termination_reason TEXT NOT NULL,
                success INTEGER NOT NULL,
                trace_id TEXT,
                created_at TEXT NOT NULL
            )
            """
        )
        connection.execute(
            f"""
            CREATE TABLE IF NOT EXISTS {self._table('approval_events')} (
                id TEXT PRIMARY KEY,
                workspace_id TEXT NOT NULL,
                pipeline_name TEXT NOT NULL,
                step_name TEXT NOT NULL,
                event_type TEXT NOT NULL,
                approved INTEGER,
                corrected INTEGER,
                trace_id TEXT,
                created_at TEXT NOT NULL
            )
            """
        )
        connection.execute(
            f"CREATE INDEX IF NOT EXISTS idx_{self._table('tool_events')}_segment ON {self._table('tool_events')} (workspace_id, task_type, tool_name, created_at DESC)"
        )
        connection.execute(
            f"CREATE INDEX IF NOT EXISTS idx_{self._table('fallback_events')}_segment ON {self._table('fallback_events')} (workspace_id, task_type, provider_key, created_at DESC)"
        )
        connection.execute(
            f"CREATE INDEX IF NOT EXISTS idx_{self._table('workflow_branch_events')}_segment ON {self._table('workflow_branch_events')} (workspace_id, pipeline_name, step_name, branch_name, created_at DESC)"
        )
        connection.execute(
            f"CREATE INDEX IF NOT EXISTS idx_{self._table('loop_events')}_segment ON {self._table('loop_events')} (workspace_id, pipeline_name, step_name, created_at DESC)"
        )
        connection.execute(
            f"CREATE INDEX IF NOT EXISTS idx_{self._table('approval_events')}_segment ON {self._table('approval_events')} (workspace_id, pipeline_name, step_name, created_at DESC)"
        )
        if current < TOOL_WORKFLOW_LEARNING_SCHEMA_VERSION:
            set_sqlite_schema_version(
                connection,
                TOOL_WORKFLOW_LEARNING_SCHEMA_KIND,
                TOOL_WORKFLOW_LEARNING_SCHEMA_VERSION,
            )
        connection.commit()

    def _placeholder(self) -> str:
        return "%s" if self.dsn else "?"

    def record_tool_call(
        self,
        *,
        tool_name: str,
        success: bool,
        duration_ms: float,
        task_type: str | None = None,
        workspace_id: str | None = None,
        trace_id: str | None = None,
        error: str | None = None,
    ) -> None:
        placeholder = self._placeholder()
        with self._connect() as connection:
            _execute(
                connection,
                f"""
                INSERT INTO {self._table('tool_events')}
                (id, workspace_id, task_type, tool_name, success, duration_ms, error, trace_id, created_at)
                VALUES ({placeholder}, {placeholder}, {placeholder}, {placeholder}, {placeholder}, {placeholder}, {placeholder}, {placeholder}, {placeholder})
                """,
                (
                    self._make_id(),
                    _segment(workspace_id),
                    _segment(task_type),
                    tool_name,
                    1 if success else 0,
                    max(0.0, float(duration_ms)),
                    error,
                    trace_id,
                    utcnow(),
                ),
                self.dsn,
            )

    def record_fallback(
        self,
        *,
        provider_key: str,
        success: bool,
        duration_ms: float,
        task_type: str | None = None,
        workspace_id: str | None = None,
        trace_id: str | None = None,
        was_primary: bool = False,
        error: str | None = None,
    ) -> None:
        placeholder = self._placeholder()
        with self._connect() as connection:
            _execute(
                connection,
                f"""
                INSERT INTO {self._table('fallback_events')}
                (id, workspace_id, task_type, provider_key, success, duration_ms, was_primary, error, trace_id, created_at)
                VALUES ({placeholder}, {placeholder}, {placeholder}, {placeholder}, {placeholder}, {placeholder}, {placeholder}, {placeholder}, {placeholder}, {placeholder})
                """,
                (
                    self._make_id(),
                    _segment(workspace_id),
                    _segment(task_type),
                    provider_key,
                    1 if success else 0,
                    max(0.0, float(duration_ms)),
                    1 if was_primary else 0,
                    error,
                    trace_id,
                    utcnow(),
                ),
                self.dsn,
            )

    def record_workflow_branch(
        self,
        *,
        pipeline_name: str,
        step_name: str,
        branch_name: str,
        success: bool,
        duration_ms: float,
        workspace_id: str | None = None,
        trace_id: str | None = None,
    ) -> None:
        placeholder = self._placeholder()
        with self._connect() as connection:
            _execute(
                connection,
                f"""
                INSERT INTO {self._table('workflow_branch_events')}
                (id, workspace_id, pipeline_name, step_name, branch_name, success, duration_ms, trace_id, created_at)
                VALUES ({placeholder}, {placeholder}, {placeholder}, {placeholder}, {placeholder}, {placeholder}, {placeholder}, {placeholder}, {placeholder})
                """,
                (
                    self._make_id(),
                    _segment(workspace_id),
                    pipeline_name,
                    step_name,
                    branch_name,
                    1 if success else 0,
                    max(0.0, float(duration_ms)),
                    trace_id,
                    utcnow(),
                ),
                self.dsn,
            )

    def record_loop_execution(
        self,
        *,
        pipeline_name: str,
        step_name: str,
        iterations: int,
        max_iterations: int,
        termination_reason: str,
        success: bool,
        workspace_id: str | None = None,
        trace_id: str | None = None,
    ) -> None:
        placeholder = self._placeholder()
        with self._connect() as connection:
            _execute(
                connection,
                f"""
                INSERT INTO {self._table('loop_events')}
                (id, workspace_id, pipeline_name, step_name, iterations, max_iterations, termination_reason, success, trace_id, created_at)
                VALUES ({placeholder}, {placeholder}, {placeholder}, {placeholder}, {placeholder}, {placeholder}, {placeholder}, {placeholder}, {placeholder}, {placeholder})
                """,
                (
                    self._make_id(),
                    _segment(workspace_id),
                    pipeline_name,
                    step_name,
                    max(0, int(iterations)),
                    max(1, int(max_iterations)),
                    termination_reason,
                    1 if success else 0,
                    trace_id,
                    utcnow(),
                ),
                self.dsn,
            )

    def record_approval_request(
        self,
        *,
        pipeline_name: str,
        step_name: str,
        workspace_id: str | None = None,
        trace_id: str | None = None,
    ) -> None:
        self._record_approval_event(
            pipeline_name=pipeline_name,
            step_name=step_name,
            workspace_id=workspace_id,
            trace_id=trace_id,
            event_type="request",
            approved=None,
            corrected=None,
        )

    def record_approval_outcome(
        self,
        *,
        pipeline_name: str,
        step_name: str,
        approved: bool,
        corrected: bool,
        workspace_id: str | None = None,
        trace_id: str | None = None,
    ) -> None:
        self._record_approval_event(
            pipeline_name=pipeline_name,
            step_name=step_name,
            workspace_id=workspace_id,
            trace_id=trace_id,
            event_type="outcome",
            approved=approved,
            corrected=corrected,
        )

    def _record_approval_event(
        self,
        *,
        pipeline_name: str,
        step_name: str,
        workspace_id: str | None,
        trace_id: str | None,
        event_type: str,
        approved: bool | None,
        corrected: bool | None,
    ) -> None:
        placeholder = self._placeholder()
        with self._connect() as connection:
            _execute(
                connection,
                f"""
                INSERT INTO {self._table('approval_events')}
                (id, workspace_id, pipeline_name, step_name, event_type, approved, corrected, trace_id, created_at)
                VALUES ({placeholder}, {placeholder}, {placeholder}, {placeholder}, {placeholder}, {placeholder}, {placeholder}, {placeholder}, {placeholder})
                """,
                (
                    self._make_id(),
                    _segment(workspace_id),
                    pipeline_name,
                    step_name,
                    event_type,
                    None if approved is None else (1 if approved else 0),
                    None if corrected is None else (1 if corrected else 0),
                    trace_id,
                    utcnow(),
                ),
                self.dsn,
            )

    def list_tool_scores(
        self,
        *,
        task_type: str | None = None,
        workspace_id: str | None = None,
    ) -> list[ToolScore]:
        placeholder = self._placeholder()
        with self._connect() as connection:
            rows = _fetchall(
                connection,
                f"""
                SELECT
                    tool_name,
                    workspace_id,
                    task_type,
                    SUM(CASE WHEN success = 1 THEN 1 ELSE 0 END) AS success_count,
                    SUM(CASE WHEN success = 0 THEN 1 ELSE 0 END) AS failure_count,
                    AVG(duration_ms) AS average_duration_ms,
                    MAX(created_at) AS last_updated
                FROM {self._table('tool_events')}
                WHERE workspace_id = {placeholder} AND task_type = {placeholder}
                GROUP BY workspace_id, task_type, tool_name
                """,
                (_segment(workspace_id), _segment(task_type)),
                self.dsn,
            )
        scores: list[ToolScore] = []
        for row in rows:
            success_count = int(_row_value(row, "success_count", 3) or 0)
            failure_count = int(_row_value(row, "failure_count", 4) or 0)
            win_rate = success_count / max(1, success_count + failure_count)
            scores.append(
                ToolScore(
                    tool_name=str(_row_value(row, "tool_name", 0)),
                    workspace_id=workspace_id,
                    task_type=task_type,
                    success_count=success_count,
                    failure_count=failure_count,
                    average_duration_ms=float(_row_value(row, "average_duration_ms", 5) or 0.0),
                    win_rate=win_rate,
                    score=_score(success_count, failure_count),
                    last_updated=str(_row_value(row, "last_updated", 6) or utcnow()),
                )
            )
        scores.sort(key=lambda item: (-item.score, -item.win_rate, item.average_duration_ms, item.tool_name))
        return scores

    def recommended_tool_names(
        self,
        tool_names: list[str],
        *,
        task_type: str | None = None,
        workspace_id: str | None = None,
    ) -> list[str]:
        score_map = {
            item.tool_name: item
            for item in self.list_tool_scores(task_type=task_type, workspace_id=workspace_id)
        }
        return sorted(
            list(tool_names),
            key=lambda name: (
                -(score_map.get(name).score if name in score_map else 0.5),
                score_map.get(name).average_duration_ms if name in score_map else 0.0,
                tool_names.index(name),
            ),
        )

    def list_fallback_scores(
        self,
        *,
        task_type: str | None = None,
        workspace_id: str | None = None,
    ) -> list[FallbackPolicyScore]:
        placeholder = self._placeholder()
        with self._connect() as connection:
            rows = _fetchall(
                connection,
                f"""
                SELECT
                    provider_key,
                    workspace_id,
                    task_type,
                    SUM(CASE WHEN success = 1 THEN 1 ELSE 0 END) AS success_count,
                    SUM(CASE WHEN success = 0 THEN 1 ELSE 0 END) AS failure_count,
                    AVG(duration_ms) AS average_duration_ms,
                    MAX(created_at) AS last_updated
                FROM {self._table('fallback_events')}
                WHERE workspace_id = {placeholder} AND task_type = {placeholder} AND was_primary = 0
                GROUP BY workspace_id, task_type, provider_key
                """,
                (_segment(workspace_id), _segment(task_type)),
                self.dsn,
            )
        scores: list[FallbackPolicyScore] = []
        for row in rows:
            success_count = int(_row_value(row, "success_count", 3) or 0)
            failure_count = int(_row_value(row, "failure_count", 4) or 0)
            win_rate = success_count / max(1, success_count + failure_count)
            scores.append(
                FallbackPolicyScore(
                    provider_key=str(_row_value(row, "provider_key", 0)),
                    workspace_id=workspace_id,
                    task_type=task_type,
                    success_count=success_count,
                    failure_count=failure_count,
                    average_duration_ms=float(_row_value(row, "average_duration_ms", 5) or 0.0),
                    win_rate=win_rate,
                    score=_score(success_count, failure_count),
                    last_updated=str(_row_value(row, "last_updated", 6) or utcnow()),
                )
            )
        scores.sort(
            key=lambda item: (-item.score, -item.win_rate, item.average_duration_ms, item.provider_key)
        )
        return scores

    def recommended_fallback_keys(
        self,
        provider_keys: list[str],
        *,
        task_type: str | None = None,
        workspace_id: str | None = None,
    ) -> list[str]:
        score_map = {
            item.provider_key: item
            for item in self.list_fallback_scores(task_type=task_type, workspace_id=workspace_id)
        }
        return sorted(
            list(provider_keys),
            key=lambda key: (
                -(score_map.get(key).score if key in score_map else 0.5),
                score_map.get(key).average_duration_ms if key in score_map else 0.0,
                provider_keys.index(key),
            ),
        )

    def list_workflow_branch_scores(
        self,
        *,
        pipeline_name: str | None = None,
        step_name: str | None = None,
        workspace_id: str | None = None,
    ) -> list[WorkflowBranchScore]:
        placeholder = self._placeholder()
        clauses = [f"workspace_id = {placeholder}"]
        params: list[Any] = [_segment(workspace_id)]
        if pipeline_name is not None:
            clauses.append(f"pipeline_name = {placeholder}")
            params.append(pipeline_name)
        if step_name is not None:
            clauses.append(f"step_name = {placeholder}")
            params.append(step_name)
        query = f"""
            SELECT
                pipeline_name,
                step_name,
                branch_name,
                SUM(CASE WHEN success = 1 THEN 1 ELSE 0 END) AS success_count,
                SUM(CASE WHEN success = 0 THEN 1 ELSE 0 END) AS failure_count,
                AVG(duration_ms) AS average_duration_ms,
                MAX(created_at) AS last_updated
            FROM {self._table('workflow_branch_events')}
            WHERE {' AND '.join(clauses)}
            GROUP BY workspace_id, pipeline_name, step_name, branch_name
        """
        with self._connect() as connection:
            rows = _fetchall(connection, query, tuple(params), self.dsn)
        scores: list[WorkflowBranchScore] = []
        for row in rows:
            success_count = int(_row_value(row, "success_count", 3) or 0)
            failure_count = int(_row_value(row, "failure_count", 4) or 0)
            scores.append(
                WorkflowBranchScore(
                    pipeline_name=str(_row_value(row, "pipeline_name", 0)),
                    step_name=str(_row_value(row, "step_name", 1)),
                    branch_name=str(_row_value(row, "branch_name", 2)),
                    workspace_id=workspace_id,
                    success_count=success_count,
                    failure_count=failure_count,
                    average_duration_ms=float(_row_value(row, "average_duration_ms", 5) or 0.0),
                    score=_score(success_count, failure_count),
                    last_updated=str(_row_value(row, "last_updated", 6) or utcnow()),
                )
            )
        scores.sort(key=lambda item: (-item.score, item.average_duration_ms, item.branch_name))
        return scores

    def recommend_loop_limit(
        self,
        pipeline_name: str,
        step_name: str,
        *,
        base_limit: int,
        workspace_id: str | None = None,
    ) -> int:
        placeholder = self._placeholder()
        with self._connect() as connection:
            rows = _fetchall(
                connection,
                f"""
                SELECT iterations, max_iterations, termination_reason
                FROM {self._table('loop_events')}
                WHERE workspace_id = {placeholder} AND pipeline_name = {placeholder} AND step_name = {placeholder}
                ORDER BY created_at DESC
                LIMIT 50
                """,
                (_segment(workspace_id), pipeline_name, step_name),
                self.dsn,
            )
        if not rows:
            return max(1, int(base_limit))
        avg_iterations = sum(int(_row_value(row, "iterations", 0) or 0) for row in rows) / max(1, len(rows))
        max_hits = sum(1 for row in rows if _row_value(row, "termination_reason", 2) == "max_iterations")
        if max_hits >= max(2, len(rows) // 3):
            return max(int(base_limit), min(int(base_limit * 2), math.ceil(avg_iterations + 1)))
        target = max(1, math.ceil(avg_iterations + 1))
        return min(max(1, int(base_limit)), target)

    def list_approval_scores(
        self,
        *,
        pipeline_name: str | None = None,
        step_name: str | None = None,
        workspace_id: str | None = None,
    ) -> list[ApprovalScore]:
        placeholder = self._placeholder()
        clauses = [f"workspace_id = {placeholder}"]
        params: list[Any] = [_segment(workspace_id)]
        if pipeline_name is not None:
            clauses.append(f"pipeline_name = {placeholder}")
            params.append(pipeline_name)
        if step_name is not None:
            clauses.append(f"step_name = {placeholder}")
            params.append(step_name)
        query = f"""
            SELECT
                pipeline_name,
                step_name,
                SUM(CASE WHEN event_type = 'request' THEN 1 ELSE 0 END) AS request_count,
                SUM(CASE WHEN event_type = 'outcome' AND approved = 1 THEN 1 ELSE 0 END) AS approval_count,
                SUM(CASE WHEN event_type = 'outcome' AND approved = 0 THEN 1 ELSE 0 END) AS rejection_count,
                SUM(CASE WHEN event_type = 'outcome' AND corrected = 1 THEN 1 ELSE 0 END) AS correction_count,
                MAX(created_at) AS last_updated
            FROM {self._table('approval_events')}
            WHERE {' AND '.join(clauses)}
            GROUP BY workspace_id, pipeline_name, step_name
        """
        with self._connect() as connection:
            rows = _fetchall(connection, query, tuple(params), self.dsn)
        scores: list[ApprovalScore] = []
        for row in rows:
            request_count = int(_row_value(row, "request_count", 2) or 0)
            approval_count = int(_row_value(row, "approval_count", 3) or 0)
            rejection_count = int(_row_value(row, "rejection_count", 4) or 0)
            correction_count = int(_row_value(row, "correction_count", 5) or 0)
            outcomes = max(1, approval_count + rejection_count)
            scores.append(
                ApprovalScore(
                    pipeline_name=str(_row_value(row, "pipeline_name", 0)),
                    step_name=str(_row_value(row, "step_name", 1)),
                    workspace_id=workspace_id,
                    request_count=request_count,
                    approval_count=approval_count,
                    rejection_count=rejection_count,
                    correction_rate=correction_count / outcomes,
                    approval_rate=approval_count / outcomes,
                    last_updated=str(_row_value(row, "last_updated", 6) or utcnow()),
                )
            )
        scores.sort(key=lambda item: (-item.correction_rate, item.approval_rate, item.step_name))
        return scores

    def list_loop_recommendations(
        self,
        *,
        pipeline_name: str | None = None,
        workspace_id: str | None = None,
    ) -> list[LoopRecommendation]:
        placeholder = self._placeholder()
        clauses = [f"workspace_id = {placeholder}"]
        params: list[Any] = [_segment(workspace_id)]
        if pipeline_name is not None:
            clauses.append(f"pipeline_name = {placeholder}")
            params.append(pipeline_name)
        query = f"""
            SELECT
                pipeline_name,
                step_name,
                AVG(iterations) AS average_iterations,
                MAX(iterations) AS max_observed_iterations,
                SUM(CASE WHEN termination_reason = 'max_iterations' THEN 1 ELSE 0 END) AS max_hits,
                MAX(max_iterations) AS latest_max_iterations,
                COUNT(*) AS sample_size,
                MAX(created_at) AS last_updated
            FROM {self._table('loop_events')}
            WHERE {' AND '.join(clauses)}
            GROUP BY workspace_id, pipeline_name, step_name
        """
        with self._connect() as connection:
            rows = _fetchall(connection, query, tuple(params), self.dsn)
        items: list[LoopRecommendation] = []
        for row in rows:
            base_limit = int(_row_value(row, "latest_max_iterations", 5) or 1)
            items.append(
                LoopRecommendation(
                    pipeline_name=str(_row_value(row, "pipeline_name", 0)),
                    step_name=str(_row_value(row, "step_name", 1)),
                    workspace_id=workspace_id,
                    recommended_limit=self.recommend_loop_limit(
                        str(_row_value(row, "pipeline_name", 0)),
                        str(_row_value(row, "step_name", 1)),
                        base_limit=base_limit,
                        workspace_id=workspace_id,
                    ),
                    average_iterations=float(_row_value(row, "average_iterations", 2) or 0.0),
                    max_observed_iterations=int(_row_value(row, "max_observed_iterations", 3) or 0),
                    max_hits=int(_row_value(row, "max_hits", 4) or 0),
                    sample_size=int(_row_value(row, "sample_size", 6) or 0),
                    last_updated=str(_row_value(row, "last_updated", 7) or utcnow()),
                )
            )
        items.sort(key=lambda item: (-item.recommended_limit, -item.average_iterations, item.step_name))
        return items

    def _make_id(self) -> str:
        from uuid import uuid4

        return uuid4().hex


def _execute(connection: sqlite3.Connection | Any, query: str, params: tuple[Any, ...] | list[Any], use_postgresql: bool) -> None:
    if use_postgresql:
        with connection.cursor() as cursor:
            cursor.execute(query, tuple(params))
        return
    connection.execute(query, tuple(params))


def _fetchall(connection: sqlite3.Connection | Any, query: str, params: tuple[Any, ...] | list[Any], use_postgresql: bool) -> list[Any]:
    if use_postgresql:
        with connection.cursor() as cursor:
            cursor.execute(query, tuple(params))
            return list(cursor.fetchall())
    return list(connection.execute(query, tuple(params)).fetchall())


def _row_value(row: Any, name: str, index: int) -> Any:
    if row is None:
        return None
    try:
        return row[name]
    except (KeyError, TypeError, IndexError):
        pass
    try:
        return getattr(row, name)
    except AttributeError:
        pass
    try:
        return row[index]
    except (IndexError, TypeError, KeyError):
        return None


__all__ = [
    "TOOL_WORKFLOW_LEARNING_SCHEMA_KIND",
    "TOOL_WORKFLOW_LEARNING_SCHEMA_VERSION",
    "ApprovalScore",
    "FallbackPolicyScore",
    "LoopRecommendation",
    "ToolScore",
    "ToolWorkflowLearningEngine",
    "WorkflowBranchScore",
]



"""Shared dataclasses, TypedDicts, and protocols used across team mixins."""
from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Protocol, TypedDict, runtime_checkable

from logos_adk.crew import CrewProcessMode, CrewTask, CrewTaskResult, CrewTaskType
from logos_adk.crew_checkpoint_store import JSONObject
from logos_adk.types import Response


@runtime_checkable
class AgentLike(Protocol):
    """Structural protocol for objects that can participate in a Team.

    This captures the minimum interface Team requires from its agents.
    Both ``Agent`` and any duck-typed object with ``name`` and ``run()``
    satisfy this protocol (e.g. test stubs, nested Teams).
    """

    @property
    def name(self) -> str: ...

    async def run(
        self, prompt: str, *, session_id: str | None = ..., **kwargs: Any
    ) -> Response: ...


@dataclass
class _PausedCrewRun:
    run_id: str
    process: CrewProcessMode
    tasks: list[CrewTask]
    kickoff: str
    context: dict[str, str]
    task_results: list[CrewTaskResult]
    task_results_by_name: dict[str, CrewTaskResult]
    handoffs: list[dict[str, Any]]
    pending: dict[str, CrewTask]
    completed: set[str]
    pending_approval: dict[str, Any]
    max_parallel_tasks: int | None = None
    session_id: str | None = None
    kwargs: dict[str, Any] | None = None
    manager_role: str | None = None
    carried_role: str | None = None
    carried_agent: str | None = None
    last_role: str | None = None
    last_agent: str | None = None


@dataclass
class _TeamEventSubscription:
    agent_name: str
    topic: str
    handler_name: str
    target: Any
    publish_to: str | None = None


@dataclass
class CampaignState:
    """Computed campaign state derived from persisted event bus records."""

    campaign_id: str
    total_events: int
    queued_events: int
    ready_events: int
    claimed_events: int
    completed_events: int
    errored_events: int
    next_due_at: str | None = None
    last_updated_at: str | None = None


class _CrewTaskCheckpointPayload(TypedDict):
    name: str
    description: str
    task_type: CrewTaskType
    role: str | None
    agent_name: str | None
    expected_output: str | None
    depends_on: list[str]
    context: list[str]
    async_execution: bool
    approval_message: str | None
    output_key: str | None
    handoff_to_role: str | None
    handoff_to_agent: str | None
    metadata: JSONObject


class _CrewTaskResultCheckpointPayload(TypedDict):
    task_name: str
    task_type: CrewTaskType
    assigned_agent: str
    assigned_role: str | None
    prompt: str
    output: str
    status: str
    duration_ms: float
    depends_on: list[str]
    async_execution: bool
    handoff_to_role: str | None
    handoff_to_agent: str | None
    metadata: JSONObject


class _PausedCrewRunCheckpointPayload(TypedDict):
    run_id: str
    process: CrewProcessMode
    tasks: list[_CrewTaskCheckpointPayload]
    kickoff: str
    context: dict[str, str]
    task_results: list[_CrewTaskResultCheckpointPayload]
    task_results_by_name: dict[str, _CrewTaskResultCheckpointPayload]
    handoffs: list[JSONObject]
    pending: dict[str, _CrewTaskCheckpointPayload]
    completed: list[str]
    pending_approval: JSONObject
    max_parallel_tasks: int | None
    session_id: str | None
    kwargs: JSONObject
    manager_role: str | None
    carried_role: str | None
    carried_agent: str | None
    last_role: str | None
    last_agent: str | None



"""Serve-layer authentication, permissions, and workspace scoping."""
from __future__ import annotations

from dataclasses import dataclass
import json
import os
from typing import Any

from logos_adk.rbac import default_rbac_policy, expand_roles


WORKSPACE_HEADER = "X-Workspace-Id"


@dataclass
class ServePrincipal:
    """Authenticated API principal."""

    name: str
    scopes: set[str]
    workspaces: set[str]
    roles: set[str] | None = None

    def has_scope(self, scope: str) -> bool:
        return "*" in self.scopes or scope in self.scopes

    def allows_workspace(self, workspace_id: str | None) -> bool:
        if "*" in self.workspaces:
            return True
        if workspace_id is None:
            return False
        return workspace_id in self.workspaces


def _normalize_principal(token_name: str, raw: dict[str, Any] | None) -> ServePrincipal:
    data = raw or {}
    role_scopes = expand_roles(data, default_rbac_policy())
    scopes = set(data.get("scopes") or [])
    scopes.update(role_scopes)
    if not scopes:
        scopes = {"*"}
    workspaces = set(data.get("workspaces") or ["*"])
    roles = set(data.get("roles") or ([] if data.get("role") is None else [data["role"]]))
    return ServePrincipal(name=token_name, scopes=scopes, workspaces=workspaces, roles=roles or None)


def load_api_keys_config(
    *,
    api_key: str | None = None,
    api_keys: dict[str, dict[str, Any]] | None = None,
) -> dict[str, ServePrincipal]:
    """Build a token registry from explicit args and environment."""
    registry: dict[str, ServePrincipal] = {}
    if api_keys:
        for token, config in api_keys.items():
            registry[token] = _normalize_principal(config.get("name", "api-client"), config)
    env_json = os.getenv("LOGOS_ADK_API_KEYS_JSON")
    if env_json:
        payload = json.loads(env_json)
        if isinstance(payload, dict):
            for token, config in payload.items():
                if isinstance(config, dict):
                    registry[token] = _normalize_principal(config.get("name", "env-client"), config)
    if api_key and api_key not in registry:
        registry[api_key] = ServePrincipal(name="legacy-api-key", scopes={"*"}, workspaces={"*"})
    return registry


def required_scope_for_request(path: str, method: str) -> str | None:
    """Resolve the minimum scope required for a request path."""
    if path.startswith("/health"):
        return None
    if path.startswith("/learning/prompts") and method == "GET":
        return "prompt:read"
    if path.startswith("/learning/prompts") and method == "POST":
        return "prompt:write"
    if path.startswith("/learning/recommendations") and method == "GET":
        return "learning:read"
    if path.startswith("/learning/recommendations") and method == "POST":
        return "learning:write"
    if path.startswith("/learning/improvement") and method == "POST":
        return "learning:write"
    if path.startswith("/learning/improvement") and method == "GET":
        return "learning:read"
    if path.startswith("/learning/retrieval") and method == "POST":
        return "learning:write"
    if path.startswith("/learning/retrieval") and method == "GET":
        return "learning:read"
    if path.startswith("/learning/datasets"):
        return "learning:read"
    if path.startswith("/feedback") or path.startswith("/outcomes"):
        return "learning:write"
    if path.startswith("/learning"):
        return "learning:read"
    if "/quality/" in path and method == "POST":
        return "quality:run"
    if "/quality/" in path and method == "GET":
        return "quality:read"
    if path.endswith("/run") or path.endswith("/batch") or path.endswith("/stream"):
        return "agent:run"
    if path.endswith("/ingest"):
        return "knowledge:write"
    if path.endswith("/state/versions") or path.endswith("/state/export") or path.endswith("/state/shards"):
        return "state:read"
    if path.endswith("/state/import") or path.endswith("/state/rollback"):
        return "state:write"
    if path.startswith("/artifacts") and method == "GET":
        return "state:read"
    if path.startswith("/observability") or path.startswith("/metrics"):
        return "observability:read"
    return "agent:run"


def resolve_workspace_id(headers: Any, query_params: Any) -> str | None:
    """Resolve workspace id from headers or query params."""
    header_value = headers.get(WORKSPACE_HEADER)
    if header_value:
        return str(header_value)
    query_value = query_params.get("workspace_id")
    if query_value:
        return str(query_value)
    return None


def scope_session_id(session_id: str | None, workspace_id: str | None) -> str | None:
    """Prefix a session id with workspace scope."""
    if session_id is None or workspace_id is None:
        return session_id
    if session_id.startswith(f"{workspace_id}:"):
        return session_id
    return f"{workspace_id}:{session_id}"



"""Tests for diffusion backend integration."""
from __future__ import annotations

import asyncio
import base64
import json

import httpx
import pytest

from logos_adk.diffusion import Automatic1111DiffusionBackend
from logos_adk.tools.artifacts import ArtifactToolkit


def test_automatic1111_backend_parses_txt2img_response():
    png_bytes = b"fake-png-image"
    expected_b64 = base64.b64encode(png_bytes).decode("ascii")
    captured: dict[str, object] = {}

    def handler(request: httpx.Request) -> httpx.Response:
        captured["url"] = str(request.url)
        captured["body"] = request.content.decode("utf-8")
        return httpx.Response(
            200,
            json={"images": [expected_b64], "info": '{"seed": 7}'},
        )

    client = httpx.AsyncClient(transport=httpx.MockTransport(handler))
    backend = Automatic1111DiffusionBackend(base_url="http://sd.example", client=client)
    result = asyncio.run(
        backend.generate_image(
            prompt="cinematic red scooter",
            negative_prompt="blurry",
            width=768,
            height=512,
            steps=32,
            guidance_scale=8.5,
            seed=7,
            sampler_name="Euler a",
        )
    )
    asyncio.run(client.aclose())

    assert captured["url"] == "http://sd.example/sdapi/v1/txt2img"
    assert json.loads(str(captured["body"]))["prompt"] == "cinematic red scooter"
    assert result.base64_data == expected_b64
    assert result.mime_type == "image/png"
    assert result.metadata["backend"] == "automatic1111"
    assert result.metadata["seed"] == 7
    assert result.metadata["sampler_name"] == "Euler a"


def test_automatic1111_backend_errors_when_images_missing():
    def handler(request: httpx.Request) -> httpx.Response:
        return httpx.Response(200, json={"images": []})

    client = httpx.AsyncClient(transport=httpx.MockTransport(handler))
    backend = Automatic1111DiffusionBackend(client=client)

    with pytest.raises(RuntimeError, match="returned no image payload"):
        asyncio.run(backend.generate_image(prompt="poster"))

    asyncio.run(client.aclose())


def test_artifact_toolkit_generates_diffusion_image_artifact(tmp_path, monkeypatch):
    png_b64 = base64.b64encode(b"png-data").decode("ascii")
    toolkit = ArtifactToolkit(root=str(tmp_path / "artifacts"))
    tools = {tool.name: tool for tool in toolkit.get_tools()}

    class FakeBackend:
        async def generate_image(self, **kwargs):
            return type(
                "Result",
                (),
                {
                    "base64_data": png_b64,
                    "mime_type": "image/png",
                    "metadata": {"backend": "automatic1111", "parameters": kwargs},
                },
            )()

    monkeypatch.setattr(
        "logos_adk.tools.artifacts.build_diffusion_backend",
        lambda **kwargs: FakeBackend(),
    )

    created = asyncio.run(
        tools["generate_diffusion_image_artifact"](
            name="hero-shot",
            prompt="bright marketplace banner",
            width=640,
            height=360,
            model="realvis",
        )
    )
    listed = asyncio.run(tools["list_artifacts"](kind="image"))

    assert created["name"] == "hero-shot"
    assert created["kind"] == "image"
    assert created["mime_type"] == "image/png"
    assert created["metadata"]["prompt"] == "bright marketplace banner"
    assert created["metadata"]["backend"] == "automatic1111"
    assert created["metadata"]["parameters"]["model"] == "realvis"
    assert listed[0]["id"] == created["id"]



"""Team — multi-agent orchestration for Logos ADK."""
from __future__ import annotations

import logging
from typing import Any, AsyncIterator
from uuid import uuid4

from logos_adk.a2a import A2ARequest, A2AResponse
from logos_adk.errors import MaxRoundsExceeded
from logos_adk.handler import collect_handlers
from logos_adk.mailbox import Mailbox
from logos_adk.runnable import Runnable
from logos_adk.team_config import TeamConfig
from logos_adk.tools import Tool
from logos_adk.topology import resolve_topology
from logos_adk.types import Response, StreamChunk, Usage
from logos_adk.ui_events import make_ui_event

from logos_adk.team._budgeting import BudgetMixin
from logos_adk.team._checkpoint import CheckpointMixin
from logos_adk.team._crew import CrewMixin
from logos_adk.team._event_bus import EventBusMixin
from logos_adk.team._messaging import MessagingMixin
from logos_adk.team._types import (
    AgentLike,
    CampaignState as CampaignState,
    _PausedCrewRun as _PausedCrewRun,
    _TeamEventSubscription as _TeamEventSubscription,
)


class Team(
    CrewMixin,
    CheckpointMixin,
    EventBusMixin,
    BudgetMixin,
    MessagingMixin,
    Runnable,
):
    """A group of agents that collaborate to complete tasks."""

    def __init__(
        self,
        name: str,
        *,
        agents: list[AgentLike] | None = None,
        orchestrator: str | None = None,
        topology: str | dict[str, list[str]] | None = None,
        max_rounds: int = 10,
        supervisor: Any | None = None,
        process: str | None = None,
        manager_role: str | None = None,
        checkpoint_store: Any | None = None,
        event_bus: Any | None = None,
        topic_policy: dict[str, list[str]] | None = None,
    ) -> None:
        agents = agents or []
        self._config = TeamConfig(
            name=name,
            agents=agents,
            orchestrator=orchestrator,
            topology=topology,
            max_rounds=max_rounds,
            supervisor=supervisor,
            process=process,
            manager_role=manager_role,
            topic_policy=topic_policy,
        )
        self._agents: dict[str, AgentLike] = {}
        self._mailboxes: dict[str, Mailbox] = {}
        self._pending_a2a_responses: dict[str, A2AResponse] = {}
        self._paused_crew_runs: dict[str, _PausedCrewRun] = {}
        self._checkpoint_store = checkpoint_store
        self._event_bus = event_bus
        self._topic_subscriptions: list[_TeamEventSubscription] = []
        self._team_budget_usd: float | None = None
        self._team_budget_tokens: int | None = None
        self._task_budget_usd: float | None = None
        self._task_budget_tokens: int | None = None
        self._budget_threshold_ratio: float = 0.8
        self._fail_on_budget_exhausted: bool = False

        # Register agents and detect duplicates
        for agent in agents:
            if agent.name in self._agents:
                raise ValueError(
                    f"duplicate agent name '{agent.name}' in team '{name}'"
                )
            self._agents[agent.name] = agent
            self._mailboxes[agent.name] = Mailbox()
            # Set backref for Agent instances; nested Teams don't have _team
            if hasattr(agent, '_team'):
                agent._team = self

        # Resolve orchestrator
        if orchestrator is not None and orchestrator not in self._agents:
            raise ValueError(
                f"Orchestrator '{orchestrator}' not in agents: "
                f"{list(self._agents.keys())}"
            )
        self._orchestrator_name: str | None = orchestrator or (
            agents[0].name if agents else None
        )

        # Collect @handler methods per agent
        self._handlers: dict[str, dict[type, Any]] = {}
        for agent in agents:
            self._handlers[agent.name] = collect_handlers(agent)

        # Resolve topology
        if agents:
            self._topology = resolve_topology(
                topology, list(self._agents.keys())
            )
        else:
            self._topology: dict[str, set[str]] = {}
        self._topic_policy = self._resolve_topic_policy(topic_policy)
        self._restore_persistent_tools()
        self._register_builtin_scheduled_task_subscriptions()
        for agent in agents:
            self._register_agent_topic_subscriptions(agent)

    @property
    def name(self) -> str:
        return self._config.name

    def _restore_persistent_tools(self) -> None:
        try:
            from logos_adk.tool_synthesis import GeneratedToolStore

            GeneratedToolStore().attach_tools_for_team(self)
        except (ImportError, AttributeError):
            # tool_synthesis is an optional module; skip if unavailable
            return
        except Exception:
            logging.getLogger(__name__).debug(
                "Failed to restore persistent tools for team '%s'",
                self._config.name,
                exc_info=True,
            )
            return

    def artifact_sandbox(self, *, root: str | None = None) -> Team:
        """Attach the shared artifact sandbox toolkit to every team member."""
        for agent in self._agents.values():
            if hasattr(agent, "artifact_sandbox"):
                agent.artifact_sandbox(root=root)
        return self

    def temporal_autonomy(self) -> Team:
        """Attach temporal scheduling tools to every team member."""
        for agent in self._agents.values():
            if hasattr(agent, "temporal_autonomy"):
                agent.temporal_autonomy()
        return self

    @classmethod
    def from_definition(cls, config: TeamConfig) -> Team:
        """Build a Team from a typed TeamConfig."""
        from logos_adk.agent import Agent
        from logos_adk.config_loader import build_agent_from_definition
        from logos_adk.errors import ConfigError

        agents = [build_agent_from_definition(agent_config, default_cls=Agent) for agent_config in config.agents]
        event_bus = cls._build_event_bus_from_config(config.event_bus)
        team = cls(
            config.name,
            agents=agents,
            orchestrator=config.orchestrator,
            topology=config.topology,
            max_rounds=config.max_rounds,
            supervisor=config.supervisor,
            process=config.process,
            manager_role=config.manager_role,
            event_bus=event_bus,
            topic_policy=config.topic_policy,
        )
        for subscription in config.subscriptions:
            try:
                team.subscribe(
                    subscription.agent_name,
                    subscription.topic,
                    handler=subscription.handler,
                    publish_to=subscription.publish_to,
                    name=subscription.name,
                )
            except Exception as exc:
                raise ConfigError(
                    f"Failed to register subscription '{subscription.agent_name}:{subscription.topic}': {exc}"
                ) from exc
        return team

    @classmethod
    def from_config(cls, path: str) -> Team:
        """Load a Team from a YAML config file."""
        from logos_adk.team_config_parser import load_team_config

        team = cls.from_definition(load_team_config(path))
        for agent in team._agents.values():
            if hasattr(agent, "_config_path"):
                agent._config_path = path
        return team

    def get_agent(self, name: str) -> AgentLike | None:
        """Get a registered agent by name."""
        return self._agents.get(name)

    def get_agents_by_role(self, role: str) -> list[AgentLike]:
        """Return all agents assigned to a Crew-style role."""
        return [
            agent
            for agent in self._agents.values()
            if getattr(getattr(agent, "_config", None), "role", None) == role
        ]

    def get_agent_by_role(self, role: str) -> AgentLike | None:
        """Return the first agent assigned to a Crew-style role."""
        matches = self.get_agents_by_role(role)
        return matches[0] if matches else None

    def roster(self) -> list[dict[str, str]]:
        """Return team roster metadata for role-aware orchestration."""
        roster: list[dict[str, str]] = []
        for agent in self._agents.values():
            profile = getattr(agent, "crew_profile", {})
            if not profile:
                continue
            roster.append({"name": agent.name, **profile})
        return roster

    # -- execution -----------------------------------------------------------

    async def kickoff(
        self,
        prompt: str,
        *,
        role: str | None = None,
        session_id: str | None = None,
        **kwargs: Any,
    ) -> Response:
        """Crew-style kickoff that can route directly by role."""
        effective_kwargs = self._apply_budget_kwargs(kwargs)
        if role is not None:
            target = self.get_agent_by_role(role)
            if target is None:
                raise ValueError(f"Role '{role}' is not assigned to any agent in team '{self.name}'")
            return await target.run(prompt, session_id=session_id, **effective_kwargs)
        if self.roster():
            prompt = self._role_aware_prompt(prompt)
        return await self.run(prompt, session_id=session_id, **effective_kwargs)

    async def delegate(
        self,
        task: str,
        *,
        role: str | None = None,
        to_agent: str | None = None,
        session_id: str | None = None,
        **kwargs: Any,
    ) -> Response:
        """Delegate a task directly to an agent or to the first agent with a given role."""
        effective_kwargs = self._apply_budget_kwargs(kwargs)
        if to_agent is None:
            if role is None:
                raise ValueError("delegate() requires either role or to_agent")
            agent = self.get_agent_by_role(role)
            if agent is None:
                raise ValueError(f"Role '{role}' is not assigned to any agent in team '{self.name}'")
        else:
            agent = self.get_agent(to_agent)
            if agent is None:
                raise ValueError(f"Agent '{to_agent}' is not registered in team '{self.name}'")
        return await agent.run(task, session_id=session_id, **effective_kwargs)

    async def run(
        self, prompt: str, *, session_id: str | None = None, **kwargs: Any
    ) -> Response:
        """Run the team: send prompt to orchestrator and return its response."""
        if not self._agents:
            raise ValueError(f"Team '{self.name}' has no agents")

        orchestrator = self._agents[self._orchestrator_name]
        response = await orchestrator.run(
            prompt,
            session_id=session_id,
            **self._apply_budget_kwargs(kwargs),
        )
        return response

    def _has_pending_messages(self) -> bool:
        """Check if any agent mailbox has pending messages."""
        return any(not mb.empty() for mb in self._mailboxes.values())

    async def run_rounds(self) -> Response:
        """Process mailbox messages in rounds until all mailboxes are empty.

        Raises:
            MaxRoundsExceeded: If max_rounds is exceeded and messages remain.
        """
        last_response: Response | None = None

        for _ in range(self._config.max_rounds):
            any_activity = False

            for agent_name in list(self._agents.keys()):
                mailbox = self._mailboxes[agent_name]
                messages = mailbox.drain()
                if not messages:
                    continue
                any_activity = True
                agent = self._agents[agent_name]
                handlers = self._handlers.get(agent_name, {})
                for msg in messages:
                    if isinstance(msg, A2AResponse):
                        self._pending_a2a_responses[msg.request_id] = msg
                        last_response = Response(
                            content=msg.content,
                            tool_calls=[],
                            usage=Usage(0, 0),
                            raw={"a2a_response": self._serialize_a2a_response(msg)},
                        )
                        continue
                    if isinstance(msg, A2ARequest):
                        handler = handlers.get(A2ARequest)
                        if handler is not None:
                            result = handler(msg)
                            if hasattr(result, '__await__'):
                                await result
                        else:
                            prompt = self._build_a2a_prompt(msg)
                            response = await agent.run(prompt)
                            if msg.expects_response:
                                await self.send(
                                    agent_name,
                                    msg.from_agent,
                                    A2AResponse(
                                        from_agent=agent_name,
                                        to_agent=msg.from_agent,
                                        content=response.content,
                                        conversation_id=msg.conversation_id,
                                        metadata={"task_id": msg.task_id},
                                        request_id=msg.task_id,
                                        status="completed",
                                    ),
                                )
                            last_response = response
                        continue
                    # Dispatch to @handler if message type matches
                    handler = handlers.get(type(msg))
                    if handler is not None:
                        result = handler(msg)
                        if hasattr(result, '__await__'):
                            await result
                    else:
                        content = msg if isinstance(msg, str) else str(msg)
                        last_response = await agent.run(content)

            event_response, processed_events = await self._run_event_round()
            if processed_events > 0:
                any_activity = True
                if event_response is not None:
                    last_response = event_response

            if not any_activity:
                break

        if self._has_pending_messages():
            raise MaxRoundsExceeded(
                f"Team '{self.name}' exceeded max_rounds={self._config.max_rounds}"
            )

        if last_response is None:
            last_response = Response(
                content="", tool_calls=[], usage=Usage(0, 0), raw={}
            )
        return last_response

    # -- streaming -----------------------------------------------------------

    async def stream(
        self, prompt: str, *, session_id: str | None = None, **kwargs: Any
    ) -> AsyncIterator[StreamChunk]:
        """Stream the team execution."""
        if not self._agents:
            raise ValueError(f"Team '{self.name}' has no agents")

        orchestrator = self._agents[self._orchestrator_name]
        async for chunk in orchestrator.stream(prompt, session_id=session_id, **kwargs):
            yield chunk

    async def stream_events(
        self,
        prompt: str,
        *,
        session_id: str | None = None,
        correlation_id: str | None = None,
        **kwargs: Any,
    ) -> AsyncIterator[dict[str, Any]]:
        run_id = uuid4().hex
        sequence = 0
        trace_id = kwargs.get("_trace_id")
        effective_correlation_id = correlation_id or session_id or run_id
        yield make_ui_event(
            event="graph_started",
            event_type="team",
            graph_name=self.name,
            graph_run_id=run_id,
            sequence=sequence,
            session_id=session_id,
            correlation_id=effective_correlation_id,
            trace_id=trace_id,
            node_name=self.name,
            status="running",
        )
        sequence += 1
        orchestrator = self._agents[self._orchestrator_name]
        final_event: dict[str, Any] | None = None
        async for event in orchestrator.stream_events(
            prompt,
            session_id=session_id,
            correlation_id=effective_correlation_id,
            **kwargs,
        ):
            payload = dict(event)
            payload.setdefault("metadata", {})
            payload["metadata"]["team_name"] = self.name
            if payload.get("event") == "graph_started":
                continue
            if payload.get("event") == "graph_completed":
                final_event = payload
                continue
            yield payload
        yield make_ui_event(
            event="graph_completed",
            event_type="team",
            graph_name=self.name,
            graph_run_id=run_id,
            sequence=sequence,
            session_id=session_id,
            correlation_id=effective_correlation_id,
            trace_id=trace_id,
            node_name=self.name,
            status=(final_event or {}).get("status") or "completed",
            output=(final_event or {}).get("output"),
            metadata=(final_event or {}).get("metadata"),
        )

    # -- tool conversion -----------------------------------------------------

    def as_tool(
        self, *, name: str | None = None, description: str | None = None
    ) -> Tool:
        """Convert this team into a Tool that can be used by other agents."""
        tool_name = name or self.name
        tool_desc = description or f"Delegate to the '{self.name}' team"
        team = self

        async def _team_tool(prompt: str) -> str:
            response = await team.run(prompt)
            return response.content

        return Tool(fn=_team_tool, name=tool_name, description=tool_desc)

    def _role_aware_prompt(self, prompt: str) -> str:
        roster_lines = []
        for entry in self.roster():
            line = f"- {entry['name']}: role={entry.get('role', '')}"
            if entry.get("goal"):
                line += f"; goal={entry['goal']}"
            if entry.get("backstory"):
                line += f"; backstory={entry['backstory']}"
            roster_lines.append(line)
        return (
            "Team roster:\n"
            + "\n".join(roster_lines)
            + "\n\nTask:\n"
            + prompt
        )



import pytest

from logos_adk.agent import Agent
from logos_adk.providers.mock import MockProvider
from logos_adk.team import Team
from logos_adk.tools import Tool
from logos_adk.types import Response, Usage


def _make_agent(name: str, text: str = "ok") -> Agent:
    provider = MockProvider(responses=[
        Response(content=text, tool_calls=[], usage=Usage(1, 1), raw={}),
    ])
    return Agent(name, model=provider)


class TestTeamAsTool:
    def test_as_tool_returns_tool(self):
        a = _make_agent("a")
        team = Team("research", agents=[a])
        t = team.as_tool()
        assert isinstance(t, Tool)

    def test_as_tool_name_defaults_to_team_name(self):
        a = _make_agent("a")
        team = Team("research", agents=[a])
        t = team.as_tool()
        assert t.name == "research"

    def test_as_tool_custom_name(self):
        a = _make_agent("a")
        team = Team("research", agents=[a])
        t = team.as_tool(name="do_research")
        assert t.name == "do_research"

    def test_as_tool_custom_description(self):
        a = _make_agent("a")
        team = Team("research", agents=[a])
        t = team.as_tool(description="Research things")
        assert t.description == "Research things"

    @pytest.mark.asyncio
    async def test_as_tool_executes_team(self):
        a = _make_agent("worker", "team result")
        team = Team("t", agents=[a])
        t = team.as_tool()
        result = await t(prompt="do work")
        assert result == "team result"

    def test_as_tool_schema_has_prompt_param(self):
        a = _make_agent("a")
        team = Team("t", agents=[a])
        t = team.as_tool()
        schema = t.schema
        assert "prompt" in schema["parameters"]["properties"]
        assert schema["parameters"]["required"] == ["prompt"]



"""Policy learning mechanisms."""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from logos_adk.learning.store import LearningStore

from .models import PolicyDecision
from .utils import _mean


@dataclass
class TrainingExport:
    """Exported training data for offline analysis."""

    workspace_id: str | None
    agent_name: str | None
    examples: list[dict[str, Any]]
    generated_at: str = field(default_factory=lambda: __import__('datetime').datetime.now(__import__('datetime').UTC).isoformat())
    metadata: dict[str, Any] = field(default_factory=dict)


class PolicyLearner:
    """Decision layer trained from existing runtime learning engines."""

    def __init__(
        self,
        store: "LearningStore",
        *,
        reward_model: Any | None = None,
        model_routing: Any | None = None,
        retrieval_learning: Any | None = None,
        tool_workflow: Any | None = None,
    ) -> None:
        """Initialize policy learner.
        
        Args:
            store: LearningStore instance
            reward_model: Optional reward model for scoring
            model_routing: Optional model routing engine
            retrieval_learning: Optional retrieval learning engine
            tool_workflow: Optional tool workflow learning engine
        """
        self.store = store
        self.reward_model = reward_model
        self.model_routing = model_routing
        self.retrieval_learning = retrieval_learning
        self.tool_workflow = tool_workflow

    def learn_policy(
        self,
        *,
        workspace_id: str | None = None,
        task_type: str | None = None,
        user_segment: str | None = None,
        limit: int = 500,
    ) -> PolicyDecision:
        """Learn policy decision from historical data."""
        runs = self.store.list_runs(
            workspace_id=workspace_id,
            agent_name=None,
            limit=limit,
        )
        
        if not runs:
            return self._default_policy(workspace_id, task_type, user_segment)
        
        reward_values = []
        model_counts: dict[str, int] = {}
        tool_counts: dict[str, int] = {}
        
        for run in runs:
            if self.reward_model:
                feedback = self.store.list_feedback(run_id=run.id)
                outcomes = self.store.list_outcomes(run_id=run.id)
                score = self.reward_model.score_run(run, feedback=feedback, outcomes=outcomes)
                reward_values.append(score.score)
            
            if run.model:
                model_counts[run.model] = model_counts.get(run.model, 0) + 1
            
            if run.tool_calls:
                for tool_call in run.tool_calls:
                    tool_counts[tool_call.get("name", "unknown")] = tool_counts.get(tool_call.get("name", "unknown"), 0) + 1
        
        avg_reward = _mean(reward_values, default=0.5)
        
        best_model = max(model_counts, key=model_counts.get) if model_counts else None
        preferred_tools = sorted(tool_counts, key=tool_counts.get, reverse=True)[:5] if tool_counts else []
        
        return PolicyDecision(
            workspace_id=workspace_id,
            task_type=task_type,
            user_segment=user_segment,
            model_key=best_model,
            preferred_tools=preferred_tools,
            confidence=min(1.0, len(runs) / 100.0),
            average_reward=round(avg_reward, 4),
            rationale=self._build_rationale(runs, reward_values, model_counts, tool_counts),
            metrics={
                "sample_size": len(runs),
                "model_counts": model_counts,
                "tool_counts": tool_counts,
            },
        )

    def _default_policy(
        self,
        workspace_id: str | None,
        task_type: str | None,
        user_segment: str | None,
    ) -> PolicyDecision:
        """Return default policy when insufficient data."""
        return PolicyDecision(
            workspace_id=workspace_id,
            task_type=task_type,
            user_segment=user_segment,
            confidence=0.0,
            average_reward=0.5,
            rationale=["Insufficient data for policy learning"],
            metrics={"sample_size": 0},
        )

    def _build_rationale(
        self,
        runs: list[Any],
        reward_values: list[float],
        model_counts: dict[str, int],
        tool_counts: dict[str, int],
    ) -> list[str]:
        """Build policy rationale from data."""
        rationale = []
        
        if reward_values:
            avg_reward = _mean(reward_values, default=0.5)
            if avg_reward >= 0.7:
                rationale.append(f"High average reward ({avg_reward:.2f}) indicates good policy")
            elif avg_reward <= 0.3:
                rationale.append(f"Low average reward ({avg_reward:.2f}) suggests policy issues")
        
        if model_counts:
            best_model = max(model_counts, key=model_counts.get)
            rationale.append(f"Model {best_model} most frequently used ({model_counts[best_model]} runs)")
        
        if tool_counts:
            best_tool = max(tool_counts, key=tool_counts.get)
            rationale.append(f"Tool {best_tool} most successful ({tool_counts[best_tool]} uses)")
        
        return rationale

    def export_training_data(
        self,
        *,
        workspace_id: str | None = None,
        agent_name: str | None = None,
        limit: int = 500,
    ) -> TrainingExport:
        """Export training data for offline analysis."""
        from logos_adk.learning.dataset import DatasetBuilder
        
        builder = DatasetBuilder(self.store)
        examples = builder.build(
            workspace_id=workspace_id,
            agent_name=agent_name,
            limit=limit,
            include_unreviewed=False,
        )
        
        export_examples = []
        for example in examples:
            if self.reward_model:
                score = self.reward_model.score_example(example)
                export_examples.append({
                    "run_id": example.run_id,
                    "input_text": example.input_text,
                    "output_text": example.output_text,
                    "reward_score": score.score,
                    "reward_components": [
                        {"name": c.name, "value": c.raw_value, "weight": c.weight}
                        for c in score.components
                    ],
                    "metadata": example.metadata,
                })
        
        return TrainingExport(
            workspace_id=workspace_id,
            agent_name=agent_name,
            examples=export_examples,
            metadata={
                "sample_size": len(export_examples),
                "export_type": "policy_training",
            },
        )


__all__ = ["PolicyLearner", "TrainingExport"]



"""Reflection memory for lessons learned."""

from __future__ import annotations

import json
import os
import sqlite3
from pathlib import Path

from logos_adk.persistence import (
    get_postgresql_schema_version,
    get_sqlite_schema_version,
    set_postgresql_schema_version,
    set_sqlite_schema_version,
)
from logos_adk.postgres_utils import import_psycopg, resolve_database_url

from .models import IMPROVEMENT_SCHEMA_KIND, IMPROVEMENT_SCHEMA_VERSION, ReflectionLesson


class ReflectionMemory:
    """Persistent structured memory for lessons learned."""

    def __init__(
        self,
        path: str | None = None,
        table_name: str = "reflection_lessons",
        dsn: str | None = None,
    ) -> None:
        """Initialize reflection memory.

        Args:
            path: SQLite database path
            table_name: Table name for lessons
            dsn: PostgreSQL DSN (optional)
        """
        self.path = path or os.getenv(
            "LOGOS_ADK_REFLECTION_DB_PATH", ".logos_adk/reflection_memory.db"
        )
        self.dsn = resolve_database_url(
            dsn or os.getenv("LOGOS_ADK_REFLECTION_DSN"),
            env_vars=("LOGOS_ADK_REFLECTION_DSN", "LOGOS_ADK_DATABASE_URL", "DATABASE_URL"),
        )
        self.table_name = table_name

    def _connect(self):
        """Create database connection."""
        if self.dsn:
            connection = import_psycopg().connect(self.dsn)
            self._ensure_schema(connection)
            return connection
        db_path = Path(self.path)
        if db_path.parent and not db_path.parent.exists():
            db_path.parent.mkdir(parents=True, exist_ok=True)
        connection = sqlite3.connect(self.path)
        connection.row_factory = sqlite3.Row
        self._ensure_schema(connection)
        return connection

    def _ensure_schema(self, connection) -> None:
        """Ensure database schema is up to date."""
        if self.dsn:
            self._ensure_postgresql_schema(connection)
        else:
            self._ensure_sqlite_schema(connection)
        connection.commit()

    def _ensure_postgresql_schema(self, connection) -> None:
        """Ensure PostgreSQL schema."""
        current = get_postgresql_schema_version(connection, IMPROVEMENT_SCHEMA_KIND) or 0
        with connection.cursor() as cursor:
            cursor.execute(
                f"""
                CREATE TABLE IF NOT EXISTS {self.table_name} (
                    id TEXT PRIMARY KEY,
                    workspace_id TEXT,
                    agent_name TEXT,
                    category TEXT NOT NULL,
                    lesson TEXT NOT NULL,
                    rationale TEXT NOT NULL,
                    suggested_action TEXT NOT NULL,
                    confidence DOUBLE PRECISION NOT NULL,
                    source_run_ids_json JSONB NOT NULL,
                    metadata_json JSONB NOT NULL,
                    created_at TEXT NOT NULL
                )
                """
            )
            cursor.execute(
                f"""
                CREATE INDEX IF NOT EXISTS idx_{self.table_name}_lookup
                ON {self.table_name} (workspace_id, agent_name, category, created_at DESC)
                """
            )
        if current < IMPROVEMENT_SCHEMA_VERSION:
            set_postgresql_schema_version(
                connection, IMPROVEMENT_SCHEMA_KIND, IMPROVEMENT_SCHEMA_VERSION
            )

    def _ensure_sqlite_schema(self, connection) -> None:
        """Ensure SQLite schema."""
        current = get_sqlite_schema_version(connection, IMPROVEMENT_SCHEMA_KIND) or 0
        connection.execute(
            f"""
            CREATE TABLE IF NOT EXISTS {self.table_name} (
                id TEXT PRIMARY KEY,
                workspace_id TEXT,
                agent_name TEXT,
                category TEXT NOT NULL,
                lesson TEXT NOT NULL,
                rationale TEXT NOT NULL,
                suggested_action TEXT NOT NULL,
                confidence REAL NOT NULL,
                source_run_ids_json TEXT NOT NULL,
                metadata_json TEXT NOT NULL,
                created_at TEXT NOT NULL
            )
            """
        )
        connection.execute(
            f"CREATE INDEX IF NOT EXISTS idx_{self.table_name}_lookup ON {self.table_name} (workspace_id, agent_name, category, created_at DESC)"
        )
        if current < IMPROVEMENT_SCHEMA_VERSION:
            set_sqlite_schema_version(
                connection, IMPROVEMENT_SCHEMA_KIND, IMPROVEMENT_SCHEMA_VERSION
            )

    def save_lesson(self, lesson: ReflectionLesson) -> ReflectionLesson:
        """Save lesson to memory."""
        with self._connect() as connection:
            if self.dsn:
                with connection.cursor() as cursor:
                    cursor.execute(
                        f"""
                        INSERT INTO {self.table_name}
                        (id, workspace_id, agent_name, category, lesson, rationale, suggested_action,
                         confidence, source_run_ids_json, metadata_json, created_at)
                        VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s::jsonb, %s::jsonb, %s)
                        ON CONFLICT(id) DO UPDATE SET
                            workspace_id = excluded.workspace_id,
                            agent_name = excluded.agent_name,
                            category = excluded.category,
                            lesson = excluded.lesson,
                            rationale = excluded.rationale,
                            suggested_action = excluded.suggested_action,
                            confidence = excluded.confidence,
                            source_run_ids_json = excluded.source_run_ids_json,
                            metadata_json = excluded.metadata_json,
                            created_at = excluded.created_at
                        """,
                        (
                            lesson.id,
                            lesson.workspace_id,
                            lesson.agent_name,
                            lesson.category,
                            lesson.lesson,
                            lesson.rationale,
                            lesson.suggested_action,
                            lesson.confidence,
                            json.dumps(lesson.source_run_ids),
                            json.dumps(lesson.metadata),
                            lesson.created_at,
                        ),
                    )
            else:
                connection.execute(
                    f"""
                    INSERT OR REPLACE INTO {self.table_name}
                    (id, workspace_id, agent_name, category, lesson, rationale, suggested_action,
                     confidence, source_run_ids_json, metadata_json, created_at)
                    VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
                    """,
                    (
                        lesson.id,
                        lesson.workspace_id,
                        lesson.agent_name,
                        lesson.category,
                        lesson.lesson,
                        lesson.rationale,
                        lesson.suggested_action,
                        lesson.confidence,
                        json.dumps(lesson.source_run_ids),
                        json.dumps(lesson.metadata),
                        lesson.created_at,
                    ),
                )
        return lesson

    def list_lessons(
        self,
        *,
        workspace_id: str | None = None,
        agent_name: str | None = None,
        category: str | None = None,
        limit: int = 50,
    ) -> list[ReflectionLesson]:
        """List lessons with filters."""
        clauses = []
        params = []
        if workspace_id is not None:
            placeholders = "%s" if self.dsn else "?"
            clauses.append(f"workspace_id = {placeholders}")
            params.append(workspace_id)
        if agent_name is not None:
            placeholders = "%s" if self.dsn else "?"
            clauses.append(f"agent_name = {placeholders}")
            params.append(agent_name)
        if category is not None:
            placeholders = "%s" if self.dsn else "?"
            clauses.append(f"category = {placeholders}")
            params.append(category)
        where_sql = f"WHERE {' AND '.join(clauses)}" if clauses else ""

        limit_placeholder = "%s" if self.dsn else "?"
        params.append(limit)

        with self._connect() as connection:
            if self.dsn:
                with connection.cursor() as cursor:
                    cursor.execute(
                        f"""
                        SELECT * FROM {self.table_name}
                        {where_sql}
                        ORDER BY created_at DESC
                        LIMIT {limit_placeholder}
                        """,
                        params,
                    )
                    rows = cursor.fetchall()
            else:
                rows = connection.execute(
                    f"""
                    SELECT * FROM {self.table_name}
                    {where_sql}
                    ORDER BY created_at DESC
                    LIMIT {limit_placeholder}
                    """,
                    params,
                ).fetchall()

        return [self._row_to_lesson(row) for row in rows]

    def get_lesson(self, lesson_id: str) -> ReflectionLesson | None:
        """Get lesson by ID."""
        with self._connect() as connection:
            row = connection.execute(
                f"SELECT * FROM {self.table_name} WHERE id = ?",
                (lesson_id,),
            ).fetchone()
        return self._row_to_lesson(row) if row else None

    def delete_lesson(self, lesson_id: str) -> bool:
        """Delete lesson by ID."""
        with self._connect() as connection:
            result = connection.execute(
                f"DELETE FROM {self.table_name} WHERE id = ?",
                (lesson_id,),
            )
            return result.rowcount > 0

    def _row_to_lesson(self, row) -> ReflectionLesson:
        """Convert database row to lesson."""
        if hasattr(row, "keys"):
            data = dict(row)
        else:
            data = {
                "id": row[0],
                "workspace_id": row[1],
                "agent_name": row[2],
                "category": row[3],
                "lesson": row[4],
                "rationale": row[5],
                "suggested_action": row[6],
                "confidence": row[7],
                "source_run_ids_json": row[8],
                "metadata_json": row[9],
                "created_at": row[10],
            }

        source_run_ids = data["source_run_ids_json"]
        if isinstance(source_run_ids, str):
            source_run_ids = json.loads(source_run_ids) if source_run_ids else []
        elif source_run_ids is None:
            source_run_ids = []

        metadata = data["metadata_json"]
        if isinstance(metadata, str):
            metadata = json.loads(metadata) if metadata else {}
        elif metadata is None:
            metadata = {}

        return ReflectionLesson(
            id=str(data["id"]),
            workspace_id=data["workspace_id"],
            agent_name=data["agent_name"],
            category=str(data["category"]),
            lesson=str(data["lesson"]),
            rationale=str(data["rationale"]),
            suggested_action=str(data["suggested_action"]),
            confidence=float(data["confidence"]),
            source_run_ids=source_run_ids,
            metadata=metadata,
            created_at=str(data["created_at"]),
        )

    def search_lessons(
        self,
        query: str,
        *,
        workspace_id: str | None = None,
        limit: int = 20,
    ) -> list[ReflectionLesson]:
        """Search lessons by text query."""
        search_pattern = f"%{query}%"
        with self._connect() as connection:
            rows = connection.execute(
                f"""
                SELECT * FROM {self.table_name}
                WHERE (workspace_id = ? OR ? IS NULL)
                  AND (lesson LIKE ? OR rationale LIKE ? OR suggested_action LIKE ?)
                ORDER BY confidence DESC, created_at DESC
                LIMIT ?
                """,
                (workspace_id, workspace_id, search_pattern, search_pattern, search_pattern, limit),
            ).fetchall()
        return [self._row_to_lesson(row) for row in rows]


__all__ = ["ReflectionMemory"]



"""Database schema management for retrieval learning."""
from __future__ import annotations

import sqlite3
from typing import Any

from logos_adk.persistence import (
    get_postgresql_schema_version,
    get_sqlite_schema_version,
    set_postgresql_schema_version,
    set_sqlite_schema_version,
)

from .engine import RETRIEVAL_LEARNING_SCHEMA_KIND, RETRIEVAL_LEARNING_SCHEMA_VERSION


def ensure_schema(engine: Any, connection: sqlite3.Connection | Any) -> None:
    """Ensure database schema is up to date.
    
    Args:
        engine: RetrievalLearningEngine instance
        connection: Database connection
    """
    if engine.dsn:
        _ensure_postgresql_schema(engine, connection)
    else:
        _ensure_sqlite_schema(engine, connection)


def _ensure_postgresql_schema(engine: Any, connection: Any) -> None:
    """Create PostgreSQL schema."""
    connection.execute(
        f"""
        CREATE TABLE IF NOT EXISTS {engine._table('query_rewrites')} (
            workspace_id TEXT NOT NULL,
            task_type TEXT NOT NULL,
            user_segment TEXT NOT NULL,
            original_query TEXT NOT NULL,
            rewritten_query TEXT NOT NULL,
            success_count INTEGER NOT NULL DEFAULT 0,
            failure_count INTEGER NOT NULL DEFAULT 0,
            last_used_at TEXT NOT NULL,
            PRIMARY KEY (workspace_id, task_type, user_segment, original_query, rewritten_query)
        )
        """
    )
    connection.execute(
        f"""
        CREATE TABLE IF NOT EXISTS {engine._table('events')} (
            id TEXT PRIMARY KEY,
            trace_id TEXT,
            session_id TEXT,
            workspace_id TEXT,
            task_type TEXT,
            user_segment TEXT,
            prompt_name TEXT,
            prompt_version TEXT,
            original_query TEXT NOT NULL,
            effective_query TEXT NOT NULL,
            limit_value INTEGER NOT NULL,
            filter_json TEXT,
            chunks_json TEXT NOT NULL,
            quality_score DOUBLE PRECISION NOT NULL DEFAULT 0,
            bad_context_detected INTEGER NOT NULL DEFAULT 0,
            synced_at TEXT,
            applied_feedback_ids_json TEXT NOT NULL DEFAULT '[]',
            applied_outcome_ids_json TEXT NOT NULL DEFAULT '[]',
            created_at TEXT NOT NULL
        )
        """
    )
    connection.execute(
        f"CREATE INDEX IF NOT EXISTS idx_{engine._table('events')}_trace ON {engine._table('events')} (trace_id, created_at DESC)"
    )
    connection.execute(
        f"""
        CREATE TABLE IF NOT EXISTS {engine._table('source_scores')} (
            workspace_id TEXT NOT NULL,
            source_key TEXT NOT NULL,
            connector TEXT,
            helpful_count INTEGER NOT NULL DEFAULT 0,
            harmful_count INTEGER NOT NULL DEFAULT 0,
            accepted_count INTEGER NOT NULL DEFAULT 0,
            rejected_count INTEGER NOT NULL DEFAULT 0,
            usefulness_score DOUBLE PRECISION NOT NULL DEFAULT 0,
            last_updated TEXT NOT NULL,
            PRIMARY KEY (workspace_id, source_key)
        )
        """
    )
    connection.execute(
        f"""
        CREATE TABLE IF NOT EXISTS {engine._table('chunk_scores')} (
            workspace_id TEXT NOT NULL,
            chunk_id TEXT NOT NULL,
            document_id TEXT,
            source_key TEXT NOT NULL,
            helpful_count INTEGER NOT NULL DEFAULT 0,
            harmful_count INTEGER NOT NULL DEFAULT 0,
            usefulness_score DOUBLE PRECISION NOT NULL DEFAULT 0,
            last_updated TEXT NOT NULL,
            PRIMARY KEY (workspace_id, chunk_id)
        )
        """
    )
    connection.execute(
        f"""
        CREATE TABLE IF NOT EXISTS {engine._table('profiles')} (
            workspace_id TEXT NOT NULL,
            task_type TEXT NOT NULL,
            user_segment TEXT NOT NULL,
            limit_value INTEGER NOT NULL,
            filter_json TEXT NOT NULL,
            success_count INTEGER NOT NULL DEFAULT 0,
            failure_count INTEGER NOT NULL DEFAULT 0,
            last_updated TEXT NOT NULL,
            PRIMARY KEY (workspace_id, task_type, user_segment, limit_value, filter_json)
        )
        """
    )
    connection.execute(
        f"""
        CREATE TABLE IF NOT EXISTS {engine._table('reindex_triggers')} (
            id TEXT PRIMARY KEY,
            workspace_id TEXT NOT NULL,
            source_key TEXT NOT NULL,
            reason TEXT NOT NULL,
            severity TEXT NOT NULL,
            occurrences INTEGER NOT NULL DEFAULT 1,
            status TEXT NOT NULL,
            suggested_chunk_size INTEGER,
            suggested_overlap INTEGER,
            metadata_json TEXT NOT NULL DEFAULT '{{}}',
            created_at TEXT NOT NULL,
            updated_at TEXT NOT NULL
        )
        """
    )
    connection.execute(
        f"CREATE INDEX IF NOT EXISTS idx_{engine._table('triggers')}_source ON {engine._table('reindex_triggers')} (workspace_id, source_key, updated_at DESC)"
    )
    connection.execute(
        f"""
        CREATE TABLE IF NOT EXISTS {engine._table('reindex_requests')} (
            id TEXT PRIMARY KEY,
            workspace_id TEXT NOT NULL,
            source_key TEXT NOT NULL,
            reason TEXT NOT NULL,
            requested_by TEXT,
            status TEXT NOT NULL,
            metadata_json TEXT NOT NULL DEFAULT '{{}}',
            created_at TEXT NOT NULL,
            updated_at TEXT NOT NULL
        )
        """
    )
    connection.execute(
        f"CREATE INDEX IF NOT EXISTS idx_{engine._table('requests')}_source ON {engine._table('reindex_requests')} (workspace_id, source_key, updated_at DESC)"
    )
    connection.execute(
        f"""
        CREATE TABLE IF NOT EXISTS {engine._table('source_documents')} (
            workspace_id TEXT NOT NULL,
            source_key TEXT NOT NULL,
            document_id TEXT NOT NULL,
            content TEXT NOT NULL,
            metadata_json TEXT NOT NULL,
            source_ref TEXT,
            updated_at TEXT NOT NULL,
            PRIMARY KEY (workspace_id, source_key, document_id)
        )
        """
    )
    recorded_version = get_postgresql_schema_version(connection, RETRIEVAL_LEARNING_SCHEMA_KIND)
    if recorded_version != RETRIEVAL_LEARNING_SCHEMA_VERSION:
        set_postgresql_schema_version(connection, RETRIEVAL_LEARNING_SCHEMA_KIND, RETRIEVAL_LEARNING_SCHEMA_VERSION)
    connection.commit()


def _ensure_sqlite_schema(engine: Any, connection: sqlite3.Connection) -> None:
    """Create SQLite schema."""
    connection.execute(
        f"""
        CREATE TABLE IF NOT EXISTS {engine._table('query_rewrites')} (
            workspace_id TEXT NOT NULL,
            task_type TEXT NOT NULL,
            user_segment TEXT NOT NULL,
            original_query TEXT NOT NULL,
            rewritten_query TEXT NOT NULL,
            success_count INTEGER NOT NULL DEFAULT 0,
            failure_count INTEGER NOT NULL DEFAULT 0,
            last_used_at TEXT NOT NULL,
            PRIMARY KEY (workspace_id, task_type, user_segment, original_query, rewritten_query)
        )
        """
    )
    connection.execute(
        f"""
        CREATE TABLE IF NOT EXISTS {engine._table('events')} (
            id TEXT PRIMARY KEY,
            trace_id TEXT,
            session_id TEXT,
            workspace_id TEXT,
            task_type TEXT,
            user_segment TEXT,
            prompt_name TEXT,
            prompt_version TEXT,
            original_query TEXT NOT NULL,
            effective_query TEXT NOT NULL,
            limit_value INTEGER NOT NULL,
            filter_json TEXT,
            chunks_json TEXT NOT NULL,
            quality_score REAL NOT NULL DEFAULT 0,
            bad_context_detected INTEGER NOT NULL DEFAULT 0,
            synced_at TEXT,
            applied_feedback_ids_json TEXT NOT NULL DEFAULT '[]',
            applied_outcome_ids_json TEXT NOT NULL DEFAULT '[]',
            created_at TEXT NOT NULL
        )
        """
    )
    connection.execute(
        f"CREATE INDEX IF NOT EXISTS idx_{engine._table('events')}_trace ON {engine._table('events')} (trace_id, created_at DESC)"
    )
    connection.execute(
        f"""
        CREATE TABLE IF NOT EXISTS {engine._table('source_scores')} (
            workspace_id TEXT NOT NULL,
            source_key TEXT NOT NULL,
            connector TEXT,
            helpful_count INTEGER NOT NULL DEFAULT 0,
            harmful_count INTEGER NOT NULL DEFAULT 0,
            accepted_count INTEGER NOT NULL DEFAULT 0,
            rejected_count INTEGER NOT NULL DEFAULT 0,
            usefulness_score REAL NOT NULL DEFAULT 0,
            last_updated TEXT NOT NULL,
            PRIMARY KEY (workspace_id, source_key)
        )
        """
    )
    connection.execute(
        f"""
        CREATE TABLE IF NOT EXISTS {engine._table('chunk_scores')} (
            workspace_id TEXT NOT NULL,
            chunk_id TEXT NOT NULL,
            document_id TEXT,
            source_key TEXT NOT NULL,
            helpful_count INTEGER NOT NULL DEFAULT 0,
            harmful_count INTEGER NOT NULL DEFAULT 0,
            usefulness_score REAL NOT NULL DEFAULT 0,
            last_updated TEXT NOT NULL,
            PRIMARY KEY (workspace_id, chunk_id)
        )
        """
    )
    connection.execute(
        f"""
        CREATE TABLE IF NOT EXISTS {engine._table('profiles')} (
            workspace_id TEXT NOT NULL,
            task_type TEXT NOT NULL,
            user_segment TEXT NOT NULL,
            limit_value INTEGER NOT NULL,
            filter_json TEXT NOT NULL,
            success_count INTEGER NOT NULL DEFAULT 0,
            failure_count INTEGER NOT NULL DEFAULT 0,
            last_updated TEXT NOT NULL,
            PRIMARY KEY (workspace_id, task_type, user_segment, limit_value, filter_json)
        )
        """
    )
    connection.execute(
        f"""
        CREATE TABLE IF NOT EXISTS {engine._table('reindex_triggers')} (
            id TEXT PRIMARY KEY,
            workspace_id TEXT NOT NULL,
            source_key TEXT NOT NULL,
            reason TEXT NOT NULL,
            severity TEXT NOT NULL,
            occurrences INTEGER NOT NULL DEFAULT 1,
            status TEXT NOT NULL,
            suggested_chunk_size INTEGER,
            suggested_overlap INTEGER,
            metadata_json TEXT NOT NULL DEFAULT '{{}}',
            created_at TEXT NOT NULL,
            updated_at TEXT NOT NULL
        )
        """
    )
    connection.execute(
        f"CREATE INDEX IF NOT EXISTS idx_{engine._table('triggers')}_source ON {engine._table('reindex_triggers')} (workspace_id, source_key, updated_at DESC)"
    )
    connection.execute(
        f"""
        CREATE TABLE IF NOT EXISTS {engine._table('reindex_requests')} (
            id TEXT PRIMARY KEY,
            workspace_id TEXT NOT NULL,
            source_key TEXT NOT NULL,
            reason TEXT NOT NULL,
            requested_by TEXT,
            status TEXT NOT NULL,
            metadata_json TEXT NOT NULL DEFAULT '{{}}',
            created_at TEXT NOT NULL,
            updated_at TEXT NOT NULL
        )
        """
    )
    connection.execute(
        f"CREATE INDEX IF NOT EXISTS idx_{engine._table('requests')}_source ON {engine._table('reindex_requests')} (workspace_id, source_key, updated_at DESC)"
    )
    connection.execute(
        f"""
        CREATE TABLE IF NOT EXISTS {engine._table('source_documents')} (
            workspace_id TEXT NOT NULL,
            source_key TEXT NOT NULL,
            document_id TEXT NOT NULL,
            content TEXT NOT NULL,
            metadata_json TEXT NOT NULL,
            source_ref TEXT,
            updated_at TEXT NOT NULL,
            PRIMARY KEY (workspace_id, source_key, document_id)
        )
        """
    )
    recorded_version = get_sqlite_schema_version(connection, RETRIEVAL_LEARNING_SCHEMA_KIND)
    if recorded_version != RETRIEVAL_LEARNING_SCHEMA_VERSION:
        set_sqlite_schema_version(connection, RETRIEVAL_LEARNING_SCHEMA_KIND, RETRIEVAL_LEARNING_SCHEMA_VERSION)
    connection.commit()



"""Logos ADK CLI."""
from __future__ import annotations

from logos_adk.agent import Agent
from logos_adk.runnable_config_loader import (
    describe_runnable_from_config,
    load_runnable_from_config,
)

from .helpers import (
    _build_continuous_eval_runner,
    _build_learning_store,
    _build_quality_manager,
)
from .main import cli, main
from .utils.animation import run_with_animation as _run_with_animation

__version__ = "0.3.0"

__all__ = [
    "cli",
    "main",
    "__version__",
    "_build_learning_store",
    "_build_quality_manager",
    "_build_continuous_eval_runner",
    "_run_with_animation",
    "Agent",
    "load_runnable_from_config",
    "describe_runnable_from_config",
]



"""PostgreSQL compatibility tests for compact learning modules."""

from __future__ import annotations

import json

import pytest

from logos_adk.learning import (
    ContinuousEvalRunner,
    ModelRoutingLearningEngine,
    PromptRecommendation,
    PromptRegistry,
    RetrievalLearningEngine,
    ToolWorkflowLearningEngine,
)
from logos_adk.rag import Chunk
from logos_adk.learning.improvement import ReflectionMemory
from logos_adk.learning.improvement.models import ReflectionLesson
from logos_adk.learning.meta_learning import InterventionType, MetaLearningEngine


class _FakeCursor:
    def __init__(self, storage):
        self.storage = storage
        self._fetchone = None
        self._fetchall = []

    def execute(self, query, params=None):
        normalized = " ".join(str(query).split())
        self._fetchone = None
        self._fetchall = []

        if normalized.startswith("INSERT INTO meta_interventions"):
            self.storage["meta"][params[0]] = {
                "intervention_id": params[0],
                "agent_name": params[1],
                "intervention_type": params[2],
                "source": params[3],
                "trace_id": params[4],
                "context": params[5],
                "cost_estimate": params[6],
                "created_at": params[7],
                "uplift": None,
                "uplift_measured_at": None,
                "status": params[8],
            }
            return

        if normalized.startswith("SELECT * FROM meta_interventions WHERE intervention_id = %s"):
            self._fetchone = self.storage["meta"].get(params[0])
            return

        if "SET uplift = %s, uplift_measured_at = %s, status = 'measured'" in normalized:
            record = self.storage["meta"].get(params[2])
            if record is not None:
                record["uplift"] = params[0]
                record["uplift_measured_at"] = params[1]
                record["status"] = "measured"
            return

        if normalized.startswith("INSERT INTO reflection_lessons"):
            self.storage["reflection"][params[0]] = {
                "id": params[0],
                "workspace_id": params[1],
                "agent_name": params[2],
                "category": params[3],
                "lesson": params[4],
                "rationale": params[5],
                "suggested_action": params[6],
                "confidence": params[7],
                "source_run_ids_json": params[8],
                "metadata_json": params[9],
                "created_at": params[10],
            }
            return

        if normalized.startswith("SELECT * FROM reflection_lessons"):
            rows = list(self.storage["reflection"].values())
            param_index = 0
            if "workspace_id = %s" in normalized:
                rows = [row for row in rows if row["workspace_id"] == params[param_index]]
                param_index += 1
            if "agent_name = %s" in normalized:
                rows = [row for row in rows if row["agent_name"] == params[param_index]]
                param_index += 1
            if "category = %s" in normalized:
                rows = [row for row in rows if row["category"] == params[param_index]]
                param_index += 1
            limit = int(params[-1]) if params else len(rows)
            rows.sort(key=lambda item: item["created_at"], reverse=True)
            self._fetchall = rows[:limit]

    def fetchone(self):
        return self._fetchone

    def fetchall(self):
        return self._fetchall

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, tb):
        return False


class _FakeConnection:
    def __init__(self):
        self.storage = {"meta": {}, "reflection": {}}
        self.commits = 0

    def cursor(self):
        return _FakeCursor(self.storage)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, tb):
        return False

    def commit(self):
        self.commits += 1

    def close(self):
        return None


class _FakeConnectionWithExecute:
    """Fake connection that supports direct execute (like SQLite)."""

    def __init__(self, storage):
        self.storage = storage
        self.commits = 0

    def execute(self, query, params=None):
        return _FakeCursor(self.storage)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, tb):
        return False

    def commit(self):
        self.commits += 1

    def close(self):
        return None


class _FakePromptRegistryCursor:
    def __init__(self, storage):
        self.storage = storage
        self._fetchone = None
        self._fetchall = []

    def execute(self, query, params=None):
        normalized = " ".join(str(query).split())
        params = tuple(params or ())
        self._fetchone = None
        self._fetchall = []

        if normalized.startswith("CREATE TABLE") or normalized.startswith("CREATE INDEX"):
            return
        if "logos_adk_schema_meta" in normalized:
            return
        if normalized.startswith("INSERT INTO prompts"):
            self.storage["prompts"][params[0]] = {
                "name": params[0],
                "description": params[1],
                "created_at": params[2],
            }
            return
        if normalized.startswith("SELECT * FROM prompts WHERE name = %s"):
            self._fetchone = self.storage["prompts"].get(params[0])
            return
        if normalized.startswith(
            "SELECT COALESCE(MAX(version), 0) AS max_version FROM prompt_versions WHERE prompt_name = %s"
        ):
            max_version = max(
                (
                    row["version"]
                    for row in self.storage["versions"].values()
                    if row["prompt_name"] == params[0]
                ),
                default=0,
            )
            self._fetchone = {"max_version": max_version}
            return
        if normalized.startswith("INSERT INTO prompt_versions"):
            self.storage["versions"][params[0]] = {
                "id": params[0],
                "prompt_name": params[1],
                "version": params[2],
                "content": params[3],
                "tags_json": json.loads(params[4]) if params[4] else [],
                "notes": params[5],
                "created_at": params[6],
                "created_by": params[7],
            }
            return
        if normalized.startswith(
            "SELECT * FROM prompt_versions WHERE prompt_name = %s AND version = %s"
        ):
            for row in self.storage["versions"].values():
                if row["prompt_name"] == params[0] and row["version"] == params[1]:
                    self._fetchone = row
                    break
            return
        if normalized.startswith(
            "SELECT * FROM prompt_versions WHERE prompt_name = %s ORDER BY version DESC"
        ):
            rows = [
                row for row in self.storage["versions"].values() if row["prompt_name"] == params[0]
            ]
            self._fetchall = sorted(rows, key=lambda item: item["version"], reverse=True)
            return
        if normalized.startswith("UPDATE prompt_deployments SET active = 0"):
            for row in self.storage["deployments"]:
                if (
                    row["prompt_name"] == params[0]
                    and row["environment"] == params[1]
                    and (row["workspace_id"] or "") == (params[2] or "")
                    and (row["task_type"] or "") == (params[3] or "")
                    and (row["user_segment"] or "") == (params[4] or "")
                ):
                    row["active"] = 0
            return
        if normalized.startswith("INSERT INTO prompt_deployments"):
            self.storage["deployments"].append(
                {
                    "id": params[0],
                    "prompt_name": params[1],
                    "version": params[2],
                    "environment": params[3],
                    "workspace_id": params[4],
                    "task_type": params[5],
                    "user_segment": params[6],
                    "active": params[7],
                    "promoted_at": params[8],
                    "metadata_json": json.loads(params[9]) if params[9] else {},
                }
            )
            return
        if normalized.startswith("SELECT * FROM prompt_deployments WHERE"):
            rows = list(self.storage["deployments"])
            param_index = 0
            if "prompt_name = %s" in normalized:
                rows = [row for row in rows if row["prompt_name"] == params[param_index]]
                param_index += 1
            if "environment = %s" in normalized:
                rows = [row for row in rows if row["environment"] == params[param_index]]
                param_index += 1
            if "active = 1" in normalized:
                rows = [row for row in rows if int(row["active"]) == 1]
            if "workspace_id = %s" in normalized:
                rows = [row for row in rows if row["workspace_id"] == params[param_index]]
                param_index += 1
            if "task_type = %s" in normalized:
                rows = [row for row in rows if row["task_type"] == params[param_index]]
                param_index += 1
            if "user_segment = %s" in normalized:
                rows = [row for row in rows if row["user_segment"] == params[param_index]]
            self._fetchall = sorted(rows, key=lambda item: item["promoted_at"], reverse=True)
            return
        if normalized.startswith("INSERT INTO prompt_experiments"):
            self.storage["experiments"][params[0]] = {
                "id": params[0],
                "prompt_name": params[1],
                "baseline_version": params[2],
                "challenger_versions_json": json.loads(params[3]) if params[3] else [],
                "metric": params[4],
                "rollout_percentage": params[5],
                "environment": params[6],
                "workspace_id": params[7],
                "task_type": params[8],
                "user_segment": params[9],
                "status": params[10],
                "created_at": params[11],
                "ended_at": params[12],
                "metadata_json": json.loads(params[13]) if params[13] else {},
            }
            return
        if normalized.startswith("SELECT * FROM prompt_experiments WHERE id = %s"):
            self._fetchone = self.storage["experiments"].get(params[0])
            return
        if normalized.startswith("SELECT * FROM prompt_experiments"):
            rows = list(self.storage["experiments"].values())
            param_index = 0
            if "prompt_name = %s" in normalized:
                rows = [row for row in rows if row["prompt_name"] == params[param_index]]
                param_index += 1
            if "status = %s" in normalized:
                rows = [row for row in rows if row["status"] == params[param_index]]
            self._fetchall = sorted(rows, key=lambda item: item["created_at"], reverse=True)
            return
        if normalized.startswith(
            "UPDATE prompt_experiments SET status = %s, ended_at = %s WHERE id = %s"
        ):
            record = self.storage["experiments"].get(params[2])
            if record is not None:
                record["status"] = params[0]
                record["ended_at"] = params[1]
            return
        if normalized.startswith("INSERT INTO prompt_recommendations"):
            self.storage["recommendations"].append(
                {
                    "id": params[0],
                    "prompt_name": params[1],
                    "kind": params[2],
                    "title": params[3],
                    "summary": params[4],
                    "proposed_patch": params[5],
                    "based_on_version": params[6],
                    "failure_counts_json": json.loads(params[7]) if params[7] else {},
                    "status": params[8],
                    "created_at": params[9],
                    "metadata_json": json.loads(params[10]) if params[10] else {},
                }
            )
            return
        if normalized.startswith("SELECT * FROM prompt_recommendations"):
            rows = list(self.storage["recommendations"])
            limit = int(params[-1]) if params else len(rows)
            if "prompt_name = %s" in normalized:
                rows = [row for row in rows if row["prompt_name"] == params[0]]
            self._fetchall = sorted(rows, key=lambda item: item["created_at"], reverse=True)[:limit]

    def fetchone(self):
        return self._fetchone

    def fetchall(self):
        return self._fetchall

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, tb):
        return False


class _FakePromptRegistryConnection:
    def __init__(self):
        self.storage = {
            "prompts": {},
            "versions": {},
            "deployments": [],
            "experiments": {},
            "recommendations": [],
        }

    def cursor(self):
        return _FakePromptRegistryCursor(self.storage)

    def commit(self):
        return None

    def close(self):
        return None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, tb):
        return False


class _FakeToolWorkflowCursor:
    def __init__(self, storage):
        self.storage = storage
        self._fetchall = []

    def execute(self, query, params=None):
        normalized = " ".join(str(query).split())
        params = tuple(params or ())
        self._fetchall = []

        if normalized.startswith("CREATE TABLE") or normalized.startswith("CREATE INDEX"):
            return
        if "logos_adk_schema_meta" in normalized:
            return
        if "INSERT INTO tool_workflow_learning_tool_events" in normalized:
            self.storage["tool_events"].append(
                {
                    "workspace_id": params[1],
                    "task_type": params[2],
                    "tool_name": params[3],
                    "success": params[4],
                    "duration_ms": params[5],
                    "created_at": params[8],
                }
            )
            return
        if "INSERT INTO tool_workflow_learning_fallback_events" in normalized:
            self.storage["fallback_events"].append(
                {
                    "workspace_id": params[1],
                    "task_type": params[2],
                    "provider_key": params[3],
                    "success": params[4],
                    "duration_ms": params[5],
                    "was_primary": params[6],
                    "created_at": params[9],
                }
            )
            return
        if "INSERT INTO tool_workflow_learning_workflow_branch_events" in normalized:
            self.storage["workflow_branch_events"].append(
                {
                    "workspace_id": params[1],
                    "pipeline_name": params[2],
                    "step_name": params[3],
                    "branch_name": params[4],
                    "success": params[5],
                    "duration_ms": params[6],
                    "created_at": params[8],
                }
            )
            return
        if "INSERT INTO tool_workflow_learning_loop_events" in normalized:
            self.storage["loop_events"].append(
                {
                    "workspace_id": params[1],
                    "pipeline_name": params[2],
                    "step_name": params[3],
                    "iterations": params[4],
                    "max_iterations": params[5],
                    "termination_reason": params[6],
                    "success": params[7],
                    "created_at": params[9],
                }
            )
            return
        if "INSERT INTO tool_workflow_learning_approval_events" in normalized:
            self.storage["approval_events"].append(
                {
                    "workspace_id": params[1],
                    "pipeline_name": params[2],
                    "step_name": params[3],
                    "event_type": params[4],
                    "approved": params[5],
                    "corrected": params[6],
                    "created_at": params[8],
                }
            )
            return
        if (
            "FROM tool_workflow_learning_tool_events" in normalized
            and "GROUP BY workspace_id, task_type, tool_name" in normalized
        ):
            rows = [
                row
                for row in self.storage["tool_events"]
                if row["workspace_id"] == params[0] and row["task_type"] == params[1]
            ]
            grouped = {}
            for row in rows:
                grouped.setdefault(row["tool_name"], []).append(row)
            self._fetchall = [
                {
                    "tool_name": name,
                    "success_count": sum(1 for item in items if item["success"] == 1),
                    "failure_count": sum(1 for item in items if item["success"] == 0),
                    "average_duration_ms": sum(item["duration_ms"] for item in items) / len(items),
                    "last_updated": max(item["created_at"] for item in items),
                }
                for name, items in grouped.items()
            ]
            return
        if (
            "FROM tool_workflow_learning_fallback_events" in normalized
            and "GROUP BY workspace_id, task_type, provider_key" in normalized
        ):
            rows = [
                row
                for row in self.storage["fallback_events"]
                if row["workspace_id"] == params[0]
                and row["task_type"] == params[1]
                and row["was_primary"] == 0
            ]
            grouped = {}
            for row in rows:
                grouped.setdefault(row["provider_key"], []).append(row)
            self._fetchall = [
                {
                    "provider_key": name,
                    "success_count": sum(1 for item in items if item["success"] == 1),
                    "failure_count": sum(1 for item in items if item["success"] == 0),
                    "average_duration_ms": sum(item["duration_ms"] for item in items) / len(items),
                    "last_updated": max(item["created_at"] for item in items),
                }
                for name, items in grouped.items()
            ]
            return
        if (
            "FROM tool_workflow_learning_workflow_branch_events" in normalized
            and "GROUP BY workspace_id, pipeline_name, step_name, branch_name" in normalized
        ):
            rows = [
                row
                for row in self.storage["workflow_branch_events"]
                if row["workspace_id"] == params[0]
            ]
            param_index = 1
            if "pipeline_name = %s" in normalized:
                rows = [row for row in rows if row["pipeline_name"] == params[param_index]]
                param_index += 1
            if "step_name = %s" in normalized:
                rows = [row for row in rows if row["step_name"] == params[param_index]]
            grouped = {}
            for row in rows:
                key = (row["pipeline_name"], row["step_name"], row["branch_name"])
                grouped.setdefault(key, []).append(row)
            self._fetchall = [
                {
                    "pipeline_name": key[0],
                    "step_name": key[1],
                    "branch_name": key[2],
                    "success_count": sum(1 for item in items if item["success"] == 1),
                    "failure_count": sum(1 for item in items if item["success"] == 0),
                    "average_duration_ms": sum(item["duration_ms"] for item in items) / len(items),
                    "last_updated": max(item["created_at"] for item in items),
                }
                for key, items in grouped.items()
            ]
            return
        if normalized.startswith(
            "SELECT iterations, max_iterations, termination_reason FROM tool_workflow_learning_loop_events"
        ):
            rows = [
                row
                for row in self.storage["loop_events"]
                if row["workspace_id"] == params[0]
                and row["pipeline_name"] == params[1]
                and row["step_name"] == params[2]
            ]
            self._fetchall = sorted(rows, key=lambda item: item["created_at"], reverse=True)[:50]
            return
        if (
            "FROM tool_workflow_learning_approval_events" in normalized
            and "GROUP BY workspace_id, pipeline_name, step_name" in normalized
        ):
            rows = [
                row for row in self.storage["approval_events"] if row["workspace_id"] == params[0]
            ]
            param_index = 1
            if "pipeline_name = %s" in normalized:
                rows = [row for row in rows if row["pipeline_name"] == params[param_index]]
                param_index += 1
            if "step_name = %s" in normalized:
                rows = [row for row in rows if row["step_name"] == params[param_index]]
            grouped = {}
            for row in rows:
                key = (row["pipeline_name"], row["step_name"])
                grouped.setdefault(key, []).append(row)
            self._fetchall = [
                {
                    "pipeline_name": key[0],
                    "step_name": key[1],
                    "request_count": sum(1 for item in items if item["event_type"] == "request"),
                    "approval_count": sum(
                        1
                        for item in items
                        if item["event_type"] == "outcome" and item["approved"] == 1
                    ),
                    "rejection_count": sum(
                        1
                        for item in items
                        if item["event_type"] == "outcome" and item["approved"] == 0
                    ),
                    "correction_count": sum(
                        1
                        for item in items
                        if item["event_type"] == "outcome" and item["corrected"] == 1
                    ),
                    "last_updated": max(item["created_at"] for item in items),
                }
                for key, items in grouped.items()
            ]
            return
        if (
            "FROM tool_workflow_learning_loop_events" in normalized
            and "GROUP BY workspace_id, pipeline_name, step_name" in normalized
        ):
            rows = [row for row in self.storage["loop_events"] if row["workspace_id"] == params[0]]
            if "pipeline_name = %s" in normalized:
                rows = [row for row in rows if row["pipeline_name"] == params[1]]
            grouped = {}
            for row in rows:
                key = (row["pipeline_name"], row["step_name"])
                grouped.setdefault(key, []).append(row)
            self._fetchall = [
                {
                    "pipeline_name": key[0],
                    "step_name": key[1],
                    "average_iterations": sum(item["iterations"] for item in items) / len(items),
                    "max_observed_iterations": max(item["iterations"] for item in items),
                    "max_hits": sum(
                        1 for item in items if item["termination_reason"] == "max_iterations"
                    ),
                    "latest_max_iterations": max(item["max_iterations"] for item in items),
                    "sample_size": len(items),
                    "last_updated": max(item["created_at"] for item in items),
                }
                for key, items in grouped.items()
            ]

    def fetchall(self):
        return self._fetchall

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, tb):
        return False


class _FakeToolWorkflowConnection:
    def __init__(self):
        self.storage = {
            "tool_events": [],
            "fallback_events": [],
            "workflow_branch_events": [],
            "loop_events": [],
            "approval_events": [],
        }

    def cursor(self):
        return _FakeToolWorkflowCursor(self.storage)

    def commit(self):
        return None

    def close(self):
        return None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, tb):
        return False


class _FakeRetrievalConnection:
    def __init__(self):
        self.query_rewrites = {}
        self.events = {}
        self.reindex_requests = {}
        self.source_documents = {}

    def execute(self, query, params=None):
        normalized = " ".join(str(query).split())
        params = tuple(params or ())
        if normalized.startswith("CREATE TABLE") or normalized.startswith("CREATE INDEX"):
            return _FakeQueryResult()
        if "logos_adk_schema_meta" in normalized:
            return _FakeQueryResult()
        if (
            "SELECT success_count, failure_count FROM retrieval_learning_query_rewrites"
            in normalized
        ):
            key = params[:5]
            row = self.query_rewrites.get(key)
            return _FakeQueryResult(one=None if row is None else row.copy())
        if normalized.startswith("INSERT INTO retrieval_learning_query_rewrites"):
            key = params[:5]
            self.query_rewrites[key] = {
                "workspace_id": params[0],
                "task_type": params[1],
                "user_segment": params[2],
                "original_query": params[3],
                "rewritten_query": params[4],
                "success_count": params[5],
                "failure_count": params[6],
                "last_used_at": params[7],
            }
            return _FakeQueryResult()
        if (
            "SELECT rewritten_query, success_count, failure_count FROM retrieval_learning_query_rewrites"
            in normalized
        ):
            rows = [
                row.copy()
                for row in self.query_rewrites.values()
                if row["workspace_id"] == params[0]
                and row["task_type"] == params[1]
                and row["user_segment"] == params[2]
                and row["original_query"] == params[3]
            ]
            rows.sort(
                key=lambda item: (
                    item["success_count"] - item["failure_count"],
                    item["success_count"],
                    item["last_used_at"],
                ),
                reverse=True,
            )
            return _FakeQueryResult(all_rows=rows)
        if normalized.startswith("SELECT * FROM retrieval_learning_query_rewrites"):
            rows = [
                row.copy()
                for row in self.query_rewrites.values()
                if row["workspace_id"] == params[0]
                and row["task_type"] == params[1]
                and row["user_segment"] == params[2]
            ]
            rows.sort(
                key=lambda item: (
                    item["success_count"] - item["failure_count"],
                    item["success_count"],
                    item["last_used_at"],
                ),
                reverse=True,
            )
            return _FakeQueryResult(all_rows=rows)
        if normalized.startswith("INSERT INTO retrieval_learning_events"):
            self.events[params[0]] = {
                "id": params[0],
                "trace_id": params[1],
                "session_id": params[2],
                "workspace_id": params[3],
                "task_type": params[4],
                "user_segment": params[5],
                "prompt_name": params[6],
                "prompt_version": params[7],
                "original_query": params[8],
                "effective_query": params[9],
                "limit_value": params[10],
                "filter_json": params[11],
                "chunks_json": params[12],
                "quality_score": params[13],
                "bad_context_detected": params[14],
                "synced_at": params[15],
                "applied_feedback_ids_json": params[16],
                "applied_outcome_ids_json": params[17],
                "created_at": params[18],
            }
            return _FakeQueryResult()
        if normalized.startswith("SELECT * FROM retrieval_learning_events"):
            rows = list(self.events.values())
            if "WHERE trace_id = ?" in query:
                rows = [row.copy() for row in rows if row["trace_id"] == params[0]]
            else:
                rows = [row.copy() for row in rows]
            rows.sort(key=lambda item: item["created_at"], reverse=True)
            return _FakeQueryResult(all_rows=rows)
        if (
            "SELECT * FROM retrieval_learning_reindex_requests WHERE workspace_id = ?" in normalized
            and "status IN ('queued', 'running')" in normalized
        ):
            rows = [
                row.copy()
                for row in self.reindex_requests.values()
                if row["workspace_id"] == params[0]
                and row["source_key"] == params[1]
                and row["status"] in {"queued", "running"}
            ]
            rows.sort(key=lambda item: item["updated_at"], reverse=True)
            return _FakeQueryResult(all_rows=rows)
        if normalized.startswith(
            "UPDATE retrieval_learning_reindex_requests SET metadata_json = ?, updated_at = ? WHERE id = ?"
        ):
            row = self.reindex_requests.get(params[2])
            if row is not None:
                row["metadata_json"] = params[0]
                row["updated_at"] = params[1]
            return _FakeQueryResult()
        if normalized.startswith("INSERT INTO retrieval_learning_reindex_requests"):
            self.reindex_requests[params[0]] = {
                "id": params[0],
                "workspace_id": params[1],
                "source_key": params[2],
                "reason": params[3],
                "requested_by": params[4],
                "status": params[5],
                "metadata_json": params[6],
                "created_at": params[7],
                "updated_at": params[8],
            }
            return _FakeQueryResult()
        if normalized.startswith("SELECT * FROM retrieval_learning_reindex_requests WHERE id = ?"):
            row = self.reindex_requests.get(params[0])
            return _FakeQueryResult(one=None if row is None else row.copy())
        if (
            "SELECT * FROM retrieval_learning_reindex_requests WHERE workspace_id = ?" in normalized
            and "ORDER BY updated_at DESC" in normalized
        ):
            rows = [
                row.copy()
                for row in self.reindex_requests.values()
                if row["workspace_id"] == params[0]
            ]
            rows.sort(key=lambda item: item["updated_at"], reverse=True)
            return _FakeQueryResult(all_rows=rows)
        if normalized.startswith("INSERT INTO retrieval_learning_source_documents"):
            key = params[:3]
            self.source_documents[key] = {
                "workspace_id": params[0],
                "source_key": params[1],
                "document_id": params[2],
                "content": params[3],
                "metadata_json": params[4],
                "source_ref": params[5],
                "updated_at": params[6],
            }
            return _FakeQueryResult()
        if normalized.startswith(
            "SELECT workspace_id, source_key, document_id FROM retrieval_learning_source_documents"
        ):
            return _FakeQueryResult(all_rows=[])
        if normalized.startswith("SELECT * FROM retrieval_learning_source_documents"):
            rows = [
                row.copy()
                for row in self.source_documents.values()
                if row["workspace_id"] == params[0]
            ]
            if len(params) > 1:
                rows = [row for row in rows if row["source_key"] == params[1]]
            rows.sort(key=lambda item: item["updated_at"], reverse=True)
            return _FakeQueryResult(all_rows=rows)
        if normalized.startswith("DELETE FROM retrieval_learning_source_documents"):
            self.source_documents.pop((params[0], params[1], params[2]), None)
            return _FakeQueryResult()
        raise AssertionError(f"Unhandled retrieval query: {normalized}")

    def cursor(self):
        return _FakeCursor({"meta": {}, "reflection": {}})

    def commit(self):
        return None

    def close(self):
        return None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, tb):
        return False


class _FakeQueryResult:
    def __init__(self, *, one=None, all_rows=None):
        self._one = one
        self._all = list(all_rows or [])

    def fetchone(self):
        return self._one

    def fetchall(self):
        return self._all


class _FakeContinuousEvalCursor:
    def __init__(self, storage):
        self.storage = storage
        self._fetchone = None
        self._fetchall = []

    def execute(self, query, params=None):
        normalized = " ".join(str(query).split())
        params = tuple(params or ())
        self._fetchone = None
        self._fetchall = []

        if normalized.startswith("INSERT INTO continuous_eval_jobs"):
            self.storage["jobs"][params[0]] = {
                "name": params[0],
                "kind": params[1],
                "agent_name": params[2],
                "workspace_id": params[3],
                "agent_filter": params[4],
                "interval_seconds": params[5],
                "limit_count": params[6],
                "include_unreviewed": params[7],
                "metadata_json": json.loads(params[8]) if params[8] else {},
                "last_run_id": params[9],
                "last_started_at": params[10],
                "last_finished_at": params[11],
            }
            return
        if normalized.startswith("DELETE FROM continuous_eval_jobs WHERE name = %s"):
            self.storage["jobs"].pop(params[0], None)
            return
        if normalized.startswith("SELECT * FROM continuous_eval_jobs"):
            rows = list(self.storage["jobs"].values())
            if params:
                rows = [row for row in rows if row["agent_name"] == params[0]]
            rows.sort(key=lambda item: item["name"])
            self._fetchall = rows
            return
        if normalized.startswith("INSERT INTO continuous_eval_incidents"):
            self.storage["incidents"][params[0]] = {
                "id": params[0],
                "job_name": params[1],
                "kind": params[2],
                "severity": params[3],
                "message": params[4],
                "created_at": params[5],
                "run_id": params[6],
                "workspace_id": params[7],
                "agent_name": params[8],
                "prompt_name": params[9],
                "experiment_id": params[10],
                "status": params[11],
                "acknowledged_at": params[12],
                "acknowledged_by": params[13],
                "resolved_at": params[14],
                "resolved_by": params[15],
                "metadata_json": json.loads(params[16]) if params[16] else {},
            }
            return
        if normalized.startswith("SELECT * FROM continuous_eval_incidents WHERE id = %s"):
            row = self.storage["incidents"].get(params[0])
            self._fetchone = None if row is None else row
            return
        if normalized.startswith("UPDATE continuous_eval_incidents SET"):
            row = self.storage["incidents"].get(params[-1])
            if row is None:
                return
            assignment_part = normalized.split("SET ", 1)[1].rsplit(" WHERE id = %s", 1)[0]
            assignments = [item.strip() for item in assignment_part.split(",")]
            for index, assignment in enumerate(assignments):
                field_name = assignment.split(" = ", 1)[0]
                row[field_name] = params[index]
            return
        if normalized.startswith("SELECT * FROM continuous_eval_incidents"):
            rows = list(self.storage["incidents"].values())
            param_index = 0
            if "kind = %s" in normalized:
                rows = [row for row in rows if row["kind"] == params[param_index]]
                param_index += 1
            if "job_name = %s" in normalized:
                rows = [row for row in rows if row["job_name"] == params[param_index]]
                param_index += 1
            if "workspace_id = %s" in normalized:
                rows = [row for row in rows if row["workspace_id"] == params[param_index]]
                param_index += 1
            if "status = %s" in normalized:
                rows = [row for row in rows if row["status"] == params[param_index]]
                param_index += 1
            limit = int(params[-1]) if params else len(rows)
            rows.sort(key=lambda item: item["created_at"], reverse=True)
            self._fetchall = rows[:limit]

    def fetchone(self):
        return self._fetchone

    def fetchall(self):
        return self._fetchall

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, tb):
        return False


class _FakeModelRoutingConnection:
    def __init__(self):
        self.events = []
        self.budgets = {}
        self.canaries = {}

    def execute(self, query, params=None):
        normalized = " ".join(str(query).split())
        params = tuple(params or ())
        if normalized.startswith("CREATE TABLE") or normalized.startswith("CREATE INDEX"):
            return _FakeQueryResult()
        if "logos_adk_schema_meta" in normalized:
            return _FakeQueryResult()
        if normalized.startswith("INSERT INTO model_routing_learning_events"):
            self.events.append(
                {
                    "id": params[0],
                    "trace_id": params[1],
                    "workspace_id": params[2],
                    "task_type": params[3],
                    "user_segment": params[4],
                    "model_key": params[5],
                    "confidence_tier": params[6],
                    "success": params[7],
                    "latency_ms": params[8],
                    "cost_usd": params[9],
                    "quality_score": params[10],
                    "was_selected": params[11],
                    "was_canary": params[12],
                    "synced_at": params[13],
                    "created_at": params[14],
                }
            )
            return _FakeQueryResult()
        if normalized.startswith("INSERT INTO model_routing_learning_budgets"):
            key = (params[0], params[1])
            self.budgets[key] = {
                "workspace_id": params[0],
                "task_type": params[1],
                "soft_limit_usd": params[2],
                "hard_limit_usd": params[3],
                "threshold_ratio": params[4],
                "period": params[5],
                "updated_at": params[6],
            }
            return _FakeQueryResult()
        if normalized.startswith("SELECT * FROM model_routing_learning_budgets"):
            return _FakeQueryResult(one=self.budgets.get((params[0], params[1])))
        if normalized.startswith("INSERT INTO model_routing_learning_canaries"):
            self.canaries[params[0]] = {
                "id": params[0],
                "model_key": params[1],
                "rollout_percentage": params[2],
                "workspace_id": params[3],
                "task_type": params[4],
                "user_segment": params[5],
                "status": params[6],
                "created_at": params[7],
                "ended_at": params[8],
            }
            return _FakeQueryResult()
        if normalized.startswith(
            "SELECT * FROM model_routing_learning_canaries WHERE status = 'active'"
        ):
            rows = [
                row.copy()
                for row in self.canaries.values()
                if row["status"] == "active"
                and row["workspace_id"] in {params[0], ""}
                and row["task_type"] in {params[1], ""}
                and row["user_segment"] in {params[2], ""}
            ]
            rows.sort(key=lambda item: item["created_at"], reverse=True)
            return _FakeQueryResult(all_rows=rows)
        if normalized.startswith("SELECT * FROM model_routing_learning_canaries"):
            rows = list(self.canaries.values())
            param_index = 0
            if "status = %s" in normalized:
                rows = [row for row in rows if row["status"] == params[param_index]]
                param_index += 1
            if "workspace_id = %s" in normalized:
                rows = [row for row in rows if row["workspace_id"] == params[param_index]]
                param_index += 1
            if "task_type = %s" in normalized:
                rows = [row for row in rows if row["task_type"] == params[param_index]]
                param_index += 1
            if "user_segment = %s" in normalized:
                rows = [row for row in rows if row["user_segment"] == params[param_index]]
            rows = [row.copy() for row in rows]
            rows.sort(key=lambda item: item["created_at"], reverse=True)
            return _FakeQueryResult(all_rows=rows)
        if (
            "SELECT COALESCE(SUM(cost_usd), 0) AS total_cost FROM model_routing_learning_events"
            in normalized
        ):
            total_cost = sum(
                float(row["cost_usd"])
                for row in self.events
                if row["workspace_id"] == params[0]
                and (len(params) == 1 or row["task_type"] == params[1])
            )
            return _FakeQueryResult(one={"total_cost": total_cost})
        if "AVG(CASE WHEN was_canary = 1 THEN quality_score END) AS canary_quality" in normalized:
            rows = [
                row
                for row in self.events
                if row["workspace_id"] == params[0]
                and (len(params) == 1 or row["task_type"] == params[1])
            ]
            canary_scores = [
                float(row["quality_score"]) for row in rows if int(row["was_canary"]) == 1
            ]
            baseline_scores = [
                float(row["quality_score"]) for row in rows if int(row["was_canary"]) == 0
            ]
            return _FakeQueryResult(
                one={
                    "canary_quality": sum(canary_scores) / len(canary_scores)
                    if canary_scores
                    else None,
                    "baseline_quality": sum(baseline_scores) / len(baseline_scores)
                    if baseline_scores
                    else None,
                }
            )
        if "GROUP BY workspace_id, task_type, model_key, confidence_tier" in normalized:
            grouped = {}
            for row in self.events:
                if row["workspace_id"] != params[0]:
                    continue
                if (
                    len(params) >= 2
                    and "task_type = %s" in normalized
                    and row["task_type"] != params[1]
                ):
                    continue
                if "confidence_tier = %s" in normalized:
                    confidence_param = params[-1]
                    if row["confidence_tier"] != confidence_param:
                        continue
                key = (row["model_key"], row["confidence_tier"])
                grouped.setdefault(key, []).append(row)
            items = []
            for (model_key, confidence_tier), rows in grouped.items():
                success_count = sum(1 for row in rows if int(row["success"]) == 1)
                failure_count = sum(1 for row in rows if int(row["success"]) == 0)
                items.append(
                    {
                        "model_key": model_key,
                        "confidence_tier": confidence_tier,
                        "success_count": success_count,
                        "failure_count": failure_count,
                        "average_latency_ms": sum(float(row["latency_ms"]) for row in rows)
                        / len(rows),
                        "average_cost_usd": sum(float(row["cost_usd"]) for row in rows) / len(rows),
                        "average_quality_score": sum(float(row["quality_score"]) for row in rows)
                        / len(rows),
                        "canary_calls": sum(1 for row in rows if int(row["was_canary"]) == 1),
                        "last_updated": max(str(row["created_at"]) for row in rows),
                    }
                )
            return _FakeQueryResult(all_rows=items)
        raise AssertionError(f"Unhandled model routing query: {normalized}")

    def commit(self):
        return None

    def close(self):
        return None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, tb):
        return False


class _FakeContinuousEvalConnection:
    def __init__(self):
        self.storage = {"jobs": {}, "incidents": {}}

    def cursor(self):
        return _FakeContinuousEvalCursor(self.storage)

    def commit(self):
        return None

    def close(self):
        return None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, tb):
        return False


def test_meta_learning_engine_supports_postgresql_backend(monkeypatch):
    connection = _FakeConnection()
    engine = MetaLearningEngine(dsn="postgresql://user:pass@localhost/meta")
    engine._connect = lambda: connection
    monkeypatch.setattr(MetaLearningEngine, "_ensure_schema", lambda self, conn: None)

    record = engine.log_intervention(
        agent_name="agent-1",
        intervention_type=InterventionType.prompt_patch,
        source="manual",
        trace_id="trace-1",
        context={"failure_types": ["hallucination"]},
        cost_estimate=0.05,
    )
    engine.measure_uplift(record.intervention_id, 0.12)

    loaded = engine.get_intervention(record.intervention_id)

    assert loaded is not None
    assert loaded.agent_name == "agent-1"
    assert loaded.context == {"failure_types": ["hallucination"]}
    assert loaded.uplift == 0.12
    assert loaded.status.value == "measured"


def test_reflection_memory_supports_postgresql_backend(monkeypatch):
    connection = _FakeConnection()
    memory = ReflectionMemory(dsn="postgresql://user:pass@localhost/reflection")
    memory._connect = lambda: connection
    monkeypatch.setattr(ReflectionMemory, "_ensure_schema", lambda self, conn: None)

    lesson = memory.save_lesson(
        ReflectionLesson(
            id="lesson-1",
            workspace_id="acme",
            agent_name="bot",
            category="retrieval_policy",
            lesson="Prefer narrower retrieval for support queries.",
            rationale="Broad recall pulled noisy chunks.",
            suggested_action="Reduce limit and tighten filters.",
            confidence=0.9,
            source_run_ids=["run-1", "run-2"],
            metadata={"signature": "bad_retrieval"},
        )
    )

    loaded = memory.list_lessons(workspace_id="acme", agent_name="bot")

    assert lesson.id is not None
    assert len(loaded) == 1
    assert loaded[0].category == "retrieval_policy"
    assert loaded[0].source_run_ids == ["run-1", "run-2"]
    assert loaded[0].metadata["signature"] == "bad_retrieval"


def test_prompt_registry_supports_postgresql_backend(monkeypatch):
    connection = _FakePromptRegistryConnection()
    registry = PromptRegistry(dsn="postgresql://user:pass@localhost/prompts")
    registry._connect = lambda: connection
    monkeypatch.setattr(PromptRegistry, "_ensure_schema", lambda self, conn: None)

    registry.ensure_prompt("support", "Support prompt")
    base = registry.create_version("support", content="baseline", tags=["v1"])
    challenger = registry.create_version("support", content="challenger")
    registry.promote(
        "support", version=base.version, workspace_id="acme", metadata={"source": "manual"}
    )
    experiment = registry.create_experiment(
        "support",
        baseline_version=base.version,
        challenger_versions=[challenger.version],
        workspace_id="acme",
        rollout_percentage=100,
    )
    recommendation = registry.save_recommendation(
        PromptRecommendation(
            id="rec-1",
            prompt_name="support",
            kind="prompt_patch",
            title="Tighten retrieval guard",
            summary="Reduce noisy retrieval scope.",
            failure_counts={"hallucination": 2},
            metadata={"priority": "high"},
        )
    )

    resolved = registry.resolve("support", workspace_id="acme", assignment_key="session-1")
    active_experiments = registry.list_experiments("support", status="active")
    listed_recommendations = registry.list_recommendations("support", limit=5)
    ended = registry.end_experiment(experiment.id, status="completed")

    assert resolved is not None
    assert resolved.prompt_name == "support"
    assert resolved.workspace_id == "acme"
    assert resolved.version in {base.version, challenger.version}
    assert active_experiments[0].id == experiment.id
    assert listed_recommendations[0].id == recommendation.id
    assert listed_recommendations[0].failure_counts["hallucination"] == 2
    assert ended.status == "completed"


def test_tool_workflow_learning_supports_postgresql_backend(monkeypatch):
    connection = _FakeToolWorkflowConnection()
    engine = ToolWorkflowLearningEngine(dsn="postgresql://user:pass@localhost/tool-workflow")
    engine._connect = lambda: connection
    monkeypatch.setattr(ToolWorkflowLearningEngine, "_ensure_schema", lambda self, conn: None)

    engine.record_tool_call(
        tool_name="search", success=True, duration_ms=12, task_type="support", workspace_id="acme"
    )
    engine.record_tool_call(
        tool_name="search", success=False, duration_ms=32, task_type="support", workspace_id="acme"
    )
    engine.record_tool_call(
        tool_name="rerank", success=True, duration_ms=18, task_type="support", workspace_id="acme"
    )
    engine.record_fallback(
        provider_key="provider:backup",
        success=True,
        duration_ms=15,
        task_type="support",
        workspace_id="acme",
    )
    engine.record_workflow_branch(
        pipeline_name="triage",
        step_name="branch",
        branch_name="fast_lane",
        success=True,
        duration_ms=8,
        workspace_id="acme",
    )
    for _ in range(3):
        engine.record_loop_execution(
            pipeline_name="triage",
            step_name="revise",
            iterations=2,
            max_iterations=4,
            termination_reason="condition_met",
            success=True,
            workspace_id="acme",
        )
    for _ in range(2):
        engine.record_approval_request(
            pipeline_name="triage", step_name="review", workspace_id="acme"
        )
    engine.record_approval_outcome(
        pipeline_name="triage",
        step_name="review",
        approved=False,
        corrected=True,
        workspace_id="acme",
    )

    tool_scores = engine.list_tool_scores(task_type="support", workspace_id="acme")
    fallback_scores = engine.list_fallback_scores(task_type="support", workspace_id="acme")
    branch_scores = engine.list_workflow_branch_scores(
        pipeline_name="triage",
        step_name="branch",
        workspace_id="acme",
    )
    approval_scores = engine.list_approval_scores(
        pipeline_name="triage",
        step_name="review",
        workspace_id="acme",
    )
    loop_recommendations = engine.list_loop_recommendations(
        pipeline_name="triage", workspace_id="acme"
    )

    assert [item.tool_name for item in tool_scores] == ["rerank", "search"]
    assert fallback_scores[0].provider_key == "provider:backup"
    assert branch_scores[0].branch_name == "fast_lane"
    assert approval_scores[0].request_count == 2
    assert approval_scores[0].rejection_count == 1
    assert loop_recommendations[0].recommended_limit >= 1


def test_retrieval_learning_supports_postgresql_backend(monkeypatch):
    connection = _FakeRetrievalConnection()
    engine = RetrievalLearningEngine(dsn="postgresql://user:pass@localhost/retrieval")
    engine._connect = lambda: connection
    monkeypatch.setattr(RetrievalLearningEngine, "_ensure_schema", lambda self, conn: None)
    monkeypatch.setattr(
        RetrievalLearningEngine, "_prune_source_documents", lambda self, conn, **kwargs: 0
    )

    engine.record_query_rewrite(
        original_query="Return Window",
        rewritten_query="refund policy",
        workspace_id="acme",
        task_type="support",
        user_segment="vip",
        succeeded=True,
    )
    rewritten = engine.rewrite_query(
        "return window",
        workspace_id="acme",
        task_type="support",
        user_segment="vip",
    )
    event = engine.record_retrieval(
        trace_id="trace-1",
        session_id="session-1",
        workspace_id="acme",
        task_type="support",
        user_segment="vip",
        prompt_name="bot",
        prompt_version="v1",
        original_query="return window",
        effective_query="refund policy",
        limit=2,
        filter={"topic": "refund"},
        chunks=[
            Chunk(
                id="chunk-1",
                document_id="doc-1",
                content="Refund policy",
                metadata={"source_key": "refund-doc", "connector": "kb"},
            )
        ],
    )
    first_request = engine.create_force_reindex_request(
        workspace_id="acme",
        source_key="refund-doc",
        requested_by="ops-a",
        metadata={"priority": "high"},
    )
    second_request = engine.create_force_reindex_request(
        workspace_id="acme",
        source_key="refund-doc",
        requested_by="ops-b",
    )
    snapshot = engine.register_source_document(
        workspace_id="acme",
        source_key="refund-doc",
        document_id="doc-1",
        content="Refund policy v2",
        metadata={"version": 2},
        source_ref="kb://refund-doc",
    )

    listed_events = engine.list_events(trace_id="trace-1")
    rewrites = engine.list_query_rewrites(
        workspace_id="acme", task_type="support", user_segment="vip"
    )
    requests = engine.list_reindex_requests(workspace_id="acme")
    documents = engine.list_source_documents(workspace_id="acme", source_key="refund-doc")

    assert rewritten == "refund policy"
    assert listed_events[0].id == event.id
    assert listed_events[0].chunks[0]["source_key"] == "refund-doc"
    assert rewrites[0].success_count == 1
    assert first_request.id == second_request.id
    assert requests[0].metadata["coalesced_count"] >= 1
    assert "ops-b" in requests[0].metadata["coalesced_requesters"]
    assert documents[0].document_id == snapshot.document_id
    assert documents[0].metadata["version"] == 2


@pytest.mark.skip(reason="Test mock issues - needs proper implementation")
def test_continuous_eval_runner_supports_postgresql_backend(monkeypatch):
    connection = _FakeContinuousEvalConnection()
    runner = ContinuousEvalRunner(
        engine=object(),  # type: ignore[arg-type]
        quality_manager=object(),  # type: ignore[arg-type]
        dsn="postgresql://user:pass@localhost/eval",
        agent_name="bot",
    )
    runner._connect = lambda: connection
    monkeypatch.setattr(ContinuousEvalRunner, "_ensure_schema", lambda self, conn: None)

    job = runner.register_regression_job(
        "prod-regression",
        agent=object(),  # type: ignore[arg-type]
        workspace_id="acme",
        agent_name="bot",
        interval_seconds=60.0,
        limit=25,
        metadata={"created_via": "pg-test"},
    )
    incident = runner._create_incident(
        job=job,
        kind="regression_threshold_breach",
        severity="high",
        message="Pass rate dropped below threshold.",
        run_id="run-1",
        prompt_name="bot",
        metadata={"drop": 0.42},
    )

    restored = runner.restore_jobs()
    listed = runner.list_incidents(limit=10, workspace_id="acme")
    acknowledged = runner.acknowledge_incident(incident.id, actor="ops", note="triaged")
    resolved = runner.resolve_incident(incident.id, actor="ops", note="handled")

    assert restored[0]["name"] == "prod-regression"
    assert restored[0]["metadata"]["created_via"] == "pg-test"
    assert listed[0]["kind"] == "regression_threshold_breach"
    assert listed[0]["metadata"]["drop"] == 0.42
    assert acknowledged is not None
    assert acknowledged["status"] == "acknowledged"
    assert acknowledged["metadata"]["operator_notes"][0]["note"] == "triaged"
    assert resolved is not None
    assert resolved["status"] == "resolved"


@pytest.mark.skip(reason="Test mock issues - needs proper implementation")
def test_continuous_eval_runner_restore_jobs_without_agent_filter_on_postgresql(monkeypatch):
    connection = _FakeContinuousEvalConnection()
    runner = ContinuousEvalRunner(
        engine=object(),  # type: ignore[arg-type]
        quality_manager=object(),  # type: ignore[arg-type]
        dsn="postgresql://user:pass@localhost/eval",
    )
    runner._connect = lambda: connection
    monkeypatch.setattr(ContinuousEvalRunner, "_ensure_schema", lambda self, conn: None)

    runner.register_regression_job(
        "prod-regression",
        agent=object(),  # type: ignore[arg-type]
        workspace_id="acme",
        agent_name="bot",
        interval_seconds=60.0,
        limit=25,
    )
    runner.default_agent_name = None
    runner.jobs.clear()
    runner._restored = False

    restored = runner.restore_jobs()

    assert restored[0]["name"] == "prod-regression"


def test_model_routing_learning_supports_postgresql_backend(monkeypatch):
    connection = _FakeModelRoutingConnection()
    engine = ModelRoutingLearningEngine(dsn="postgresql://user:pass@localhost/model-routing")
    engine._connect = lambda: connection
    monkeypatch.setattr(ModelRoutingLearningEngine, "_ensure_schema", lambda self, conn: None)

    engine.set_budget(
        workspace_id="acme",
        task_type="support",
        soft_limit_usd=0.01,
        threshold_ratio=0.5,
    )
    canary = engine.register_canary(
        model_key="MockProvider:canary-model",
        rollout_percentage=100,
        workspace_id="acme",
        task_type="support",
    )
    engine.record_route(
        model_key="MockProvider:baseline-model",
        success=True,
        latency_ms=22,
        cost_usd=0.0002,
        confidence_tier="low",
        workspace_id="acme",
        task_type="support",
        quality_score=0.15,
    )
    engine.record_route(
        model_key="MockProvider:canary-model",
        success=True,
        latency_ms=18,
        cost_usd=0.0003,
        confidence_tier="low",
        workspace_id="acme",
        task_type="support",
        quality_score=0.8,
        was_canary=True,
    )

    budget = engine.get_budget(workspace_id="acme", task_type="support")
    budget_status = engine.budget_status(workspace_id="acme", task_type="support")
    canaries = engine.list_canaries(status="active", workspace_id="acme", task_type="support")
    scores = engine.list_model_scores(
        workspace_id="acme", task_type="support", confidence_tier="low"
    )
    decision = engine.inspect_routing_decision(
        ["MockProvider:baseline-model", "MockProvider:canary-model"],
        prompt="Need a short support reply",
        assignment_key="session-1",
        workspace_id="acme",
        task_type="support",
    )

    assert budget is not None
    assert budget.soft_limit_usd == 0.01
    assert budget_status["spent_usd"] > 0.0
    assert canaries[0].id == canary.id
    assert scores[0].model_key == "MockProvider:canary-model"
    assert decision["decision"]["selected_model_key"] == "MockProvider:canary-model"
    assert decision["decision"]["used_canary"] is True
